"""Implementation of SAML 2.0 for NDG Security - ElementTree module for
ElementTree representation of SAML objects

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "23/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: etree.py 7259 2010-08-02 12:26:32Z pjkersha $"
import logging
log = logging.getLogger(__name__)
import re

try: # python 2.5
    from xml.etree import cElementTree, ElementTree
except ImportError:
    # if you've installed it yourself it comes this way
    import cElementTree, ElementTree

from ndg.saml.saml2.core import (SAMLObject, Attribute, AttributeStatement, 
                                 AuthnStatement, AuthzDecisionStatement, 
                                 Assertion, Conditions, AttributeValue, 
                                 AttributeQuery, AuthzDecisionQuery, Subject, 
                                 NameID, Issuer, Response, Status, StatusCode, 
                                 StatusMessage, StatusDetail, Advice, Action, 
                                 Evidence, DecisionType, XSStringAttributeValue) 
                             
from ndg.saml.common import SAMLVersion
from ndg.saml.common.xml import SAMLConstants
from ndg.saml.common.xml import QName as GenericQName
from ndg.saml.xml import XMLTypeParseError, UnknownAttrProfile
from ndg.saml.utils import SAMLDateTime


# Generic ElementTree Helper classes
class QName(ElementTree.QName):
    """Extend ElementTree implementation for improved attribute access support
    """ 

    # ElementTree tag is of the form {namespace}localPart.  getNs extracts the
    # namespace from within the brackets but if not found returns ''
    getNs = staticmethod(lambda tag: getattr(re.search('(?<=\{).+(?=\})', tag),
                                             'group', 
                                             str)())
                                             
    getLocalPart = staticmethod(lambda tag: tag.rsplit('}', 1)[-1])

    def __init__(self, input, tag=None, prefix=None):
        """
        @type input: basestring
        @param input: ElementTree style namespace URI + tag name -
        {namespace URI}tag - OR if tag keyword is set, the namespace URI alone
        @type tag: basestring / None
        @param tag: element tag name.  If None, input must contain the 
        namespace URI and tag name in the ElementTree form {namespace URI}tag.
        @type prefix: basestring / None
        @param prefix: namespace prefix
        """
        
        ElementTree.QName.__init__(self, input, tag=tag)
        
        if tag:
            self.namespaceURI = input
            self.localPart = tag
        else:
            # No tag provided namespace and localPart of QN must be parsed from
            # the namespace
            self.namespaceURI = QName.getNs(input)
            self.localPart = QName.getLocalPart(input)
            
        self.prefix = prefix
    
    def _getPrefix(self):
        """@return: NS prefix
        @rtype: basestring
        """
        return self.__prefix

    def _setPrefix(self, value):
        """@param value: NS prefix
        @type value: basestring
        """
        self.__prefix = value
    
    prefix = property(_getPrefix, _setPrefix, None, "Prefix")

    def _getLocalPart(self):
        """@return: NS local name
        @rtype: basestring
        """
        return self.__localPart
    
    def _setLocalPart(self, value):
        """@param value: NS local name
        @type value: basestring
        """
        self.__localPart = value
        
    localPart = property(_getLocalPart, _setLocalPart, None, "LocalPart")

    def _getNamespaceURI(self):
        """@return: NS URI
        @rtype: basestring
        """
        return self.__namespaceURI

    def _setNamespaceURI(self, value):
        """@param value: NS URI
        @type value: basestring
        """
        self.__namespaceURI = value
  
    namespaceURI = property(_getNamespaceURI, _setNamespaceURI, None, 
                            "Namespace URI'")

    def __eq__(self, qname):
        """Enable equality check for QName.  Note that prefixes don't need to
        match
        
        @type qname: ndg.security.common.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        
        @return: True if the qualified names match, False otherwise
        @rtype: bool
        """
        if not isinstance(qname, QName):
            raise TypeError('Expecting %r; got %r' % (QName, type(qname)))
                   
        # Nb. prefixes don't need to agree!         
        return (self.namespaceURI, self.localPart) == \
               (qname.namespaceURI, qname.localPart)

    def __ne__(self, qname):
        """Enable equality check for QName.  Note that prefixes don't need to
        match
        
        @type qname: ndg.security.common.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        @return: True if the qualified names don't match, False otherwise
        @rtype: bool
        """
        return not self.__eq__(qname)

    @classmethod
    def fromGeneric(cls, genericQName):
        '''Cast the generic QName type in ndg.saml.common.xml to this
        ElementTree specific implementation
        
        @param genericQName: SAML core qualified name type
        @type genericQName: ndg.saml.common.xml.QName
        @return: ElementTree specific qualified name type
        @rtype: ndg.saml.xml.etree.QName
        '''
        if not isinstance(genericQName, GenericQName):
            raise TypeError("Expecting %r for QName, got %r" % (GenericQName,
                                                        type(genericQName)))
            
        qname = cls(genericQName.namespaceURI, 
                    tag=genericQName.localPart,
                    prefix=genericQName.prefix)
        return qname
    
    
def prettyPrint(*arg, **kw):
    '''Lightweight pretty printing of ElementTree elements.  This function
    wraps the PrettyPrint class
    
    @param arg: arguments to pretty print function
    @type arg: tuple
    @param kw: keyword arguments to pretty print function
    @type kw: dict
    '''
    
    # Keep track of namespace declarations made so they're not repeated
    declaredNss = []
    
    _prettyPrint = PrettyPrint(declaredNss)
    return _prettyPrint(*arg, **kw)


class PrettyPrint(object):
    '''Class for lightweight pretty printing of ElementTree elements'''
    def __init__(self, declaredNss):
        """
        @param declaredNss: declared namespaces
        @type declaredNss: iterable of string elements
        """
        self.declaredNss = declaredNss
    
    @staticmethod
    def estrip(elem):
        '''Utility to remove unwanted leading and trailing whitespace 
        
        @param elem: ElementTree element
        @type elem: ElementTree.Element
        @return: element content with whitespace removed
        @rtype: basestring'''
        if elem is None:
            return ''
        else:
            # just in case the elem is another simple type - e.g. int - 
            # wrapper it as a string
            return str(elem).strip()
        
    def __call__(self, elem, indent='', html=0, space=' '*4):
        '''Most of the work done in this wrapped function - wrapped so that
        state can be maintained for declared namespace declarations during
        recursive calls using "declaredNss" above
        
        @param elem: ElementTree element
        @type elem: ElementTree.Element
        @param indent: set indent for output
        @type indent: basestring
        @param space: set output spacing
        @type space: basestring 
        @return: pretty print format for doc
        @rtype: basestring       
        '''  
        strAttribs = []
        for attr, attrVal in elem.attrib.items():
            nsDeclaration = ''
            
            attrNamespace = QName.getNs(attr)
            if attrNamespace:
                nsPrefix = ElementTree._namespace_map.get(attrNamespace)
                if nsPrefix is None:
                    raise KeyError('prettyPrint: missing namespace "%s" for ' 
                                   'ElementTree._namespace_map'%attrNamespace)
                
                attr = "%s:%s" % (nsPrefix, QName.getLocalPart(attr))
                
                if attrNamespace not in self.declaredNss:
                    nsDeclaration = ' xmlns:%s="%s"' % (nsPrefix,attrNamespace)
                    self.declaredNss.append(attrNamespace)
                
            strAttribs.append('%s %s="%s"' % (nsDeclaration, attr, attrVal))
            
        strAttrib = ''.join(strAttribs)
        
        namespace = QName.getNs(elem.tag)
        nsPrefix = ElementTree._namespace_map.get(namespace)
        if nsPrefix is None:
            raise KeyError('prettyPrint: missing namespace "%s" for ' 
                           'ElementTree._namespace_map' % namespace)
            
        tag = "%s:%s" % (nsPrefix, QName.getLocalPart(elem.tag))
        
        # Put in namespace declaration if one doesn't already exist
        # FIXME: namespace declaration handling is wrong for handling child
        # element scope
        if namespace in self.declaredNss:
            nsDeclaration = ''
        else:
            nsDeclaration = ' xmlns:%s="%s"' % (nsPrefix, namespace)
            self.declaredNss.append(namespace)
            
        result = '%s<%s%s%s>%s' % (indent, tag, nsDeclaration, strAttrib, 
                                   PrettyPrint.estrip(elem.text))
        
        children = len(elem)
        if children:
            for child in elem:
                declaredNss = self.declaredNss[:]
                _prettyPrint = PrettyPrint(declaredNss)
                result += '\n'+ _prettyPrint(child, indent=indent+space) 
                
            result += '\n%s%s</%s>' % (indent,
                                       PrettyPrint.estrip(child.tail),
                                       tag)
        else:
            result += '</%s>' % tag
            
        return result


# ElementTree SAML wrapper classes
class ConditionsElementTree(Conditions):
    """ElementTree based XML representation of Conditions class
    """
    
    @classmethod
    def toXML(cls, conditions):
        """Make a tree of a XML elements based on the assertion conditions
        
        @type conditions: saml.saml2.core.Conditions
        @param conditions: Assertion conditions to be represented as an 
        ElementTree Element
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        if not isinstance(conditions, Conditions):
            raise TypeError("Expecting %r type got: %r" % (Conditions,
                                                           conditions))
        
        notBeforeStr = SAMLDateTime.toString(conditions.notBefore)
        notOnOrAfterStr = SAMLDateTime.toString(conditions.notOnOrAfter)
        attrib = {
            cls.NOT_BEFORE_ATTRIB_NAME: notBeforeStr,
            cls.NOT_ON_OR_AFTER_ATTRIB_NAME: notOnOrAfterStr,
        }
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix

        for condition in conditions.conditions:
            raise NotImplementedError("Conditions list creation is not "
                                      "implemented")
                
        return elem
    
    @classmethod
    def fromXML(cls, elem):
        """Parse an ElementTree SAML Conditions element into an
        Conditions object
        
        @type elem: ElementTree.Element
        @param elem: ElementTree element containing the conditions
        @rtype: saml.saml2.core.Conditions
        @return: Conditions object
        """
        
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        localName = QName.getLocalPart(elem.tag)
        if localName != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        
        if len(elem) > 0:
            raise NotImplementedError("Conditions list parsing is not "
                                      "implemented")

        conditions = Conditions()
        notBefore = elem.attrib.get(Conditions.NOT_BEFORE_ATTRIB_NAME)
        if notBefore is not None:
            conditions.notBefore = SAMLDateTime.fromString(notBefore)
            
        notOnOrAfter = elem.attrib.get(Conditions.NOT_ON_OR_AFTER_ATTRIB_NAME)
        if notBefore is not None:
            conditions.notOnOrAfter = SAMLDateTime.fromString(notOnOrAfter)
            
        return conditions                
        
               
class AssertionElementTree(Assertion):
    """ElementTree based XML representation of Assertion class
    """
    
    @classmethod
    def toXML(cls, assertion, **attributeValueElementTreeFactoryKw):
        """Make a tree of a XML elements based on the assertion
        
        @type assertion: saml.saml2.core.Assertion
        @param assertion: Assertion to be represented as an ElementTree Element
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        
        if not isinstance(assertion, Assertion):
            raise TypeError("Expecting %r type got: %r"%(Assertion, assertion))
        
        issueInstant = SAMLDateTime.toString(assertion.issueInstant)
        attrib = {
            cls.ID_ATTRIB_NAME: assertion.id,
            cls.ISSUE_INSTANT_ATTRIB_NAME: issueInstant,
            
            # Nb. Version is a SAMLVersion instance and requires explicit cast
            cls.VERSION_ATTRIB_NAME: str(assertion.version)
        }
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix
        
        if assertion.issuer is not None:
            issuerElem = IssuerElementTree.toXML(assertion.issuer)
            elem.append(issuerElem)
        
        if assertion.subject is not None:
            subjectElem = SubjectElementTree.toXML(assertion.subject)
            elem.append(subjectElem)

        if assertion.advice:
            raise NotImplementedError("Assertion Advice creation is not "
                                      "implemented")

        if assertion.conditions is not None:
            conditionsElem = ConditionsElementTree.toXML(assertion.conditions)
            elem.append(conditionsElem)
            
        for statement in assertion.statements:
            raise NotImplementedError("Assertion Statement creation is not "
                                      "implemented")
        
        for authnStatement in assertion.authnStatements:
            raise NotImplementedError("Assertion Authentication Statement "
                                      "creation is not implemented")
        
        for authzDecisionStatement in assertion.authzDecisionStatements:
            authzDecisionStatementElem = \
                AuthzDecisionStatementElementTree.toXML(authzDecisionStatement)
            elem.append(authzDecisionStatementElem)
            
        for attributeStatement in assertion.attributeStatements:
            attributeStatementElem = AttributeStatementElementTree.toXML(
                                        attributeStatement,
                                        **attributeValueElementTreeFactoryKw)
            elem.append(attributeStatementElem)
        
        return elem

    @classmethod
    def fromXML(cls, elem, **attributeValueElementTreeFactoryKw):
        """Parse an ElementTree representation of an Assertion into an
        Assertion object
        
        @type elem: ElementTree.Element
        @param elem: ElementTree element containing the assertion
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        @rtype: saml.saml2.core.Assertion
        @return: Assertion object
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        localName = QName.getLocalPart(elem.tag)
        if localName != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        
        # Unpack attributes from top-level element
        attributeValues = []
        for attributeName in (cls.VERSION_ATTRIB_NAME,
                              cls.ISSUE_INSTANT_ATTRIB_NAME,
                              cls.ID_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                          'element' %
                                          (attributeName,
                                           cls.DEFAULT_ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
        
        assertion = cls()
        assertion.version = SAMLVersion(attributeValues[0])
        if assertion.version != SAMLVersion.VERSION_20:
            raise NotImplementedError("Parsing for %r is implemented for "
                                      "SAML version %s only; version %s is " 
                                      "not supported" % 
                                      (cls,
                                       SAMLVersion(SAMLVersion.VERSION_20),
                                       SAMLVersion(assertion.version)))
            
        assertion.issueInstant = SAMLDateTime.fromString(attributeValues[1])
        assertion.id = attributeValues[2]
        
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName == Issuer.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Issuer
                assertion.issuer = IssuerElementTree.fromXML(childElem)
                
            elif localName == Subject.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse subject
                assertion.subject = SubjectElementTree.fromXML(childElem)
                
            elif localName == Advice.DEFAULT_ELEMENT_LOCAL_NAME:
                raise NotImplementedError("Assertion Advice parsing is not "
                                          "implemented")
                
            elif localName == Conditions.DEFAULT_ELEMENT_LOCAL_NAME:
                assertion.conditions = ConditionsElementTree.fromXML(childElem)
        
            elif localName == AuthnStatement.DEFAULT_ELEMENT_LOCAL_NAME:
                raise NotImplementedError("Assertion Authentication Statement "
                                          "parsing is not implemented")
        
            elif localName == AuthzDecisionStatement.DEFAULT_ELEMENT_LOCAL_NAME:
                authzDecisionStatement = \
                    AuthzDecisionStatementElementTree.fromXML(childElem)
                assertion.authzDecisionStatements.append(authzDecisionStatement)
            
            elif localName == AttributeStatement.DEFAULT_ELEMENT_LOCAL_NAME:
                attributeStatement = AttributeStatementElementTree.fromXML(
                                        childElem,
                                        **attributeValueElementTreeFactoryKw)
                assertion.attributeStatements.append(attributeStatement)
            else:
                raise XMLTypeParseError('Assertion child element name "%s" '
                                        'not recognised' % localName)
        
        return assertion

  
class AttributeStatementElementTree(AttributeStatement):
    """ElementTree XML representation of AttributeStatement"""
    
    @classmethod
    def toXML(cls, attributeStatement, **attributeValueElementTreeFactoryKw):
        """Make a tree of a XML elements based on the attribute statement
        
        @type attributeStatement: saml.saml2.core.AttributeStatement
        @param attributeStatement: Attribute Statement to be represented as an 
        ElementTree Element
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        if not isinstance(attributeStatement, AttributeStatement):
            raise TypeError("Expecting %r type got: %r" % (AttributeStatement, 
                                                           attributeStatement))
            
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))  
        elem = ElementTree.Element(tag)
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix 

        for attribute in attributeStatement.attributes:
            # Factory enables support for multiple attribute types
            attributeElem = AttributeElementTree.toXML(attribute,
                                        **attributeValueElementTreeFactoryKw)
            elem.append(attributeElem)
        
        return elem
    
    @classmethod
    def fromXML(cls, elem, **attributeValueElementTreeFactoryKw):
        """Parse an ElementTree SAML AttributeStatement element into an
        AttributeStatement object
        
        @type elem: ElementTree.Element
        @param elem: ElementTree element containing the AttributeStatement
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: saml.saml2.core.AttributeStatement
        @return: Attribute Statement
        """
        
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        localName = QName.getLocalPart(elem.tag)
        if localName != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        
        attributeStatement = AttributeStatement()

        for childElem in elem:
            # Factory enables support for multiple attribute types
            attribute = AttributeElementTree.fromXML(childElem,
                                        **attributeValueElementTreeFactoryKw)
            attributeStatement.attributes.append(attribute)
        
        return attributeStatement

  
class AuthzDecisionStatementElementTree(AuthzDecisionStatement):
    """ElementTree XML representation of AuthzDecisionStatement"""
    
    @classmethod
    def toXML(cls, authzDecisionStatement):
        """Make a tree of a XML elements based on the authzDecision statement
        
        @type authzDecisionStatement: saml.saml2.core.AuthzDecisionStatement
        @param authzDecisionStatement: AuthzDecision Statement to be represented
        as an ElementTree Element
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        if not isinstance(authzDecisionStatement, AuthzDecisionStatement):
            raise TypeError("Expecting %r type got: %r" % 
                            (AuthzDecisionStatement, authzDecisionStatement))
          
        if not authzDecisionStatement.resource:
            raise AttributeError("Resource for AuthzDecisionStatement is not "
                                 "set")
              
        attrib = {
            cls.DECISION_ATTRIB_NAME: str(authzDecisionStatement.decision),
            cls.RESOURCE_ATTRIB_NAME: authzDecisionStatement.resource
        }
            
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))  
        elem = ElementTree.Element(tag, **attrib)
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix 

        for action in authzDecisionStatement.actions:
            # Factory enables support for multiple authzDecision types
            actionElem = ActionElementTree.toXML(action)
            elem.append(actionElem)
        
        if (authzDecisionStatement.evidence and 
            len(authzDecisionStatement.evidence.values) > 0):
            raise NotImplementedError("authzDecisionStatementElementTree does "
                                      "not currently support the Evidence type")
            
        return elem
    
    @classmethod
    def fromXML(cls, elem):
        """Parse an ElementTree SAML AuthzDecisionStatement element into an
        AuthzDecisionStatement object
        
        @type elem: ElementTree.Element
        @param elem: ElementTree element containing the AuthzDecisionStatement
        @rtype: saml.saml2.core.AuthzDecisionStatement
        @return: AuthzDecision Statement"""
        
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        localName = QName.getLocalPart(elem.tag)
        if localName != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        # Unpack attributes from top-level element
        attributeValues = []
        for attributeName in (cls.DECISION_ATTRIB_NAME,
                              cls.RESOURCE_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                 'element' %
                                 (attributeName,
                                  cls.DEFAULT_ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
        
        authzDecisionStatement = AuthzDecisionStatement()
        authzDecisionStatement.decision = DecisionType(attributeValues[0])
        authzDecisionStatement.resource = attributeValues[1]

        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == Action.DEFAULT_ELEMENT_LOCAL_NAME:
                action = ActionElementTree.fromXML(childElem)
                authzDecisionStatement.actions.append(action)
                
            elif localName == Evidence.DEFAULT_ELEMENT_LOCAL_NAME:
                raise NotImplementedError("XML parse of %s element is not "
                                          "implemented" %
                                          Evidence.DEFAULT_ELEMENT_LOCAL_NAME)
            else:
                raise XMLTypeParseError("AuthzDecisionStatement child element "
                                        "name %r not recognised" % localName)
        
        return authzDecisionStatement


class AttributeElementTree(Attribute):
    """ElementTree XML representation of SAML Attribute object.  Extend
    to make Attribute types""" 

    @classmethod
    def toXML(cls, attribute, **attributeValueElementTreeFactoryKw):
        """Make a tree of a XML elements based on the Attribute
        
        @type attribute: saml.saml2.core.Attribute
        @param attribute: Attribute to be represented as an ElementTree Element
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        if not isinstance(attribute, Attribute):
            raise TypeError("Expecting %r type got: %r"%(Attribute, attribute))
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))    
        elem = ElementTree.Element(tag)
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix 
        
        if attribute.friendlyName:
            elem.set(cls.FRIENDLY_NAME_ATTRIB_NAME, attribute.friendlyName) 
             
        if attribute.name:
            elem.set(cls.NAME_ATTRIB_NAME, attribute.name)
        
        if attribute.nameFormat:
            elem.set(cls.NAME_FORMAT_ATTRIB_NAME, attribute.nameFormat)

        for attributeValue in attribute.attributeValues:
            factory = AttributeValueElementTreeFactory(
                                        **attributeValueElementTreeFactoryKw)
            
            attributeValueElementTree = factory(attributeValue)
            
            attributeValueElem = attributeValueElementTree.toXML(attributeValue)
            elem.append(attributeValueElem)
            
        return elem
 
    @classmethod
    def fromXML(cls, elem, **attributeValueElementTreeFactoryKw):
        """Parse ElementTree element into a SAML Attribute object
        
        @type elem: ElementTree.Element
        @param elem: Attribute as ElementTree XML element
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: saml.saml2.core.Attribute
        @return: SAML Attribute
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        attribute = Attribute()
            
        # Name is mandatory in the schema
        name = elem.attrib.get(cls.NAME_ATTRIB_NAME)
        if name is None:
            raise XMLTypeParseError('No "%s" attribute found in the "%s" '
                                    'element' %
                                    (cls.NAME_ATTRIB_NAME,
                                     cls.DEFAULT_ELEMENT_LOCAL_NAME))
        attribute.name = name
            
        friendlyName = elem.attrib.get(cls.FRIENDLY_NAME_ATTRIB_NAME)
        if friendlyName is not None:
            attribute.friendlyName = friendlyName
            
        nameFormat = elem.attrib.get(cls.NAME_FORMAT_ATTRIB_NAME)    
        if nameFormat is not None:
            attribute.nameFormat = nameFormat
        
        # Factory to handle the different Attribute Value types
        factory = AttributeValueElementTreeFactory(
                                        **attributeValueElementTreeFactoryKw)

        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName != AttributeValue.DEFAULT_ELEMENT_LOCAL_NAME:
                raise XMLTypeParseError('Expecting "%s" element; found "%s"'%
                                    (AttributeValue.DEFAULT_ELEMENT_LOCAL_NAME,
                                     localName))
                            
            attributeValueElementTreeClass = factory(childElem)
            attributeValue = attributeValueElementTreeClass.fromXML(childElem)
            attribute.attributeValues.append(attributeValue)
        
        return attribute
        
    
class AttributeValueElementTreeBase(AttributeValue):
    """Base class ElementTree XML representation of SAML Attribute Value""" 
    
    @classmethod
    def toXML(cls, attributeValue):
        """Make a tree of a XML elements based on the Attribute value
        
        @type attributeValue: saml.saml2.core.AttributeValue
        @param attributeValue: Assertion to be represented as an ElementTree 
        Element
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        if not isinstance(attributeValue, AttributeValue):
            raise TypeError("Expecting %r type got: %r" % (AttributeValue, 
                                                           attributeValue))
            
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))   
        elem = ElementTree.Element(tag)
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix

        return elem


class XSStringAttributeValueElementTree(AttributeValueElementTreeBase,
                                        XSStringAttributeValue):
    """ElementTree XML representation of SAML String type Attribute Value""" 
    
    @classmethod
    def toXML(cls, attributeValue):
        """Create an XML representation of the input SAML Attribute Value 
        
        @type attributeValue: saml.saml2.core.XSStringAttributeValue
        @param attributeValue: xs:string to be represented as an ElementTree 
        Element
        @rtype: ElementTree.Element
        @return: ElementTree Element
        """
        elem = AttributeValueElementTreeBase.toXML(attributeValue)
        
        if not isinstance(attributeValue, XSStringAttributeValue):
            raise TypeError("Expecting %r type got: %r" % 
                            (XSStringAttributeValue, attributeValue)) 
        
        # Have to explicitly add namespace declaration here rather use 
        # ElementTree._namespace_map because the prefixes are used for 
        # attributes not element names        
        elem.set("%s:%s" % (SAMLConstants.XMLNS_PREFIX, 
                            SAMLConstants.XSD_PREFIX),
                 SAMLConstants.XSD_NS)
                                   
        elem.set("%s:%s" % (SAMLConstants.XMLNS_PREFIX, 
                            SAMLConstants.XSI_PREFIX),
                 SAMLConstants.XSI_NS)
        
        elem.set("%s:%s" % (SAMLConstants.XSI_PREFIX, 'type'), 
                 "%s:%s" % (SAMLConstants.XSD_PREFIX, 
                            cls.TYPE_LOCAL_NAME))

        elem.text = attributeValue.value

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree xs:string element into a SAML 
        XSStringAttributeValue object
        
        @type elem: ElementTree.Element
        @param elem: Attribute value as ElementTree XML element
        @rtype: saml.saml2.core.AttributeValue
        @return: SAML Attribute value
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        localName = QName.getLocalPart(elem.tag)
        if localName != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        # Parse the attribute type checking that it is set to the expected 
        # string type
        typeQName = QName(SAMLConstants.XSI_NS, tag='type')
        
        typeValue = elem.attrib.get(str(typeQName), '')
        typeValueLocalName = typeValue.split(':')[-1]
        if typeValueLocalName != cls.TYPE_LOCAL_NAME:
            raise XMLTypeParseError('Expecting "%s" type; got "%s"' %
                                      (cls.TYPE_LOCAL_NAME,
                                       typeValueLocalName))
        
        # Update namespace map as an XSI type has been referenced.  This will
        # ensure the correct prefix is applied if it is re-serialised.
        ElementTree._namespace_map[SAMLConstants.XSI_NS
                                   ] = SAMLConstants.XSI_PREFIX
                                      
        attributeValue = XSStringAttributeValue()
        if elem.text is not None:
            attributeValue.value = elem.text.strip()

        return attributeValue


class AttributeValueElementTreeFactory(object):
    """Class factory for AttributeValue ElementTree classes.  These classes are
    used to represent SAML Attribute value types
    
    @type toXMLTypeMap: dict
    @cvar toXMLTypeMap: mapping between SAML AttributeValue class and its 
    ElementTree handler class
    @type toSAMLTypeMap: dict
    @cvar toSAMLTypeMap: mapping between SAML AttributeValue string identifier 
    and its ElementTree handler class
    """
    toXMLTypeMap = {
        XSStringAttributeValue: XSStringAttributeValueElementTree
    }

    def xsstringMatch(elem):
        """Match function for xs:string type attribute.
        
        @type elem: ElementTree.Element
        @param elem: Attribute Value element to be checked
        @rtype: XSStringAttributeValueElementTree/None
        @return: Parsing class if this element is an xs:string Attribute Value,
        None otherwise.
        """
        # Iterate through the attributes searching for a type attribute set to
        # xs:string
        for attribName, attribVal in elem.attrib.items():
            qname = QName(attribName)
            if qname.localPart == "type":
                typeLocalName = attribVal.split(':')[-1]
                
                if typeLocalName == XSStringAttributeValue.TYPE_LOCAL_NAME:
                    return XSStringAttributeValueElementTree
                else:
                    return None
                
        # No type attribute was found for this Attribute element
        return None
        
    toSAMLTypeMap = [xsstringMatch]
    xsstringMatch = staticmethod(toSAMLTypeMap[0])
   
    def __init__(self, customToXMLTypeMap={}, customToSAMLTypeMap=[]): 
        """Set-up a SAML class to ElementTree mapping
        
        @type customToXMLTypeMap: dict
        @param customToXMLTypeMap: mapping for custom SAML AttributeValue 
        classes to their respective ElementTree based representations.  This 
        appends to self.__toXMLTypeMap
        @type customToSAMLTypeMap: dict
        @param customToSAMLTypeMap: string ID based mapping for custom SAML 
        AttributeValue classes to their respective ElementTree based 
        representations.  As with customToXMLTypeMap, this appends to
        to the respective self.__toSAMLTypeMap
        """
        self.__toXMLTypeMap = AttributeValueElementTreeFactory.toXMLTypeMap
        if not isinstance(customToXMLTypeMap, dict):
            raise TypeError('Expecting dict type for "customToXMLTypeMap"')

        for samlClass, etreeClass in customToXMLTypeMap.items(): 
            if not issubclass(samlClass, AttributeValue):
                raise TypeError("Input custom class must be derived from %r, "
                                "got %r instead" % (Attribute, samlClass))
                
            self.__toXMLTypeMap[samlClass] = etreeClass

        if not isinstance(customToSAMLTypeMap, (list, tuple)):
            raise TypeError('Expecting list or tuple type for '
                            '"customToSAMLTypeMap"')
        
        self.__toSAMLTypeMap = AttributeValueElementTreeFactory.toSAMLTypeMap[:]
        for func in customToSAMLTypeMap:
            if not callable(func):
                raise TypeError('"customToSAMLTypeMap" items must be callable')
            
        self.__toSAMLTypeMap += customToSAMLTypeMap

    def __call__(self, input):
        """Create an ElementTree object based on the Attribute class type
        passed in
        
        @type input: saml.saml2.core.AttributeValue or basestring
        @param input: pass an AttributeValue derived type or a string.  If
        an AttributeValue type, then self.__toXMLTypeMap is checked for a 
        matching AttributeValue class entry, if a string is passed, 
        self.__toSAMLTypeMap is checked for a matching string ID.  In both 
        cases, if a match is found an ElementTree class is returned which can 
        render or parse the relevant AttributeValue class
        """
        if isinstance(input, AttributeValue):
            XMLTypeClass = self.__toXMLTypeMap.get(input.__class__)
            if XMLTypeClass is None:
                raise UnknownAttrProfile("no matching XMLType class "
                                         "representation for class %r" % 
                                         input.__class__)
                
        elif ElementTree.iselement(input):
            XMLTypeClasses = []
            for matchFunc in self.__toSAMLTypeMap:
                cls = matchFunc(input)
                if cls is None:
                    continue
                elif issubclass(cls, AttributeValue):
                    XMLTypeClasses.append(cls)
                else:
                    raise TypeError("Expecting AttributeValue derived type "
                                    "for XML class; got %r" % cls)
            
            nXMLTypeClasses = len(XMLTypeClasses)
            if nXMLTypeClasses == 0:
                raise UnknownAttrProfile("no matching XMLType class "
                                         "representation for SAML "
                                         "AttributeValue type %r" % input)
            elif nXMLTypeClasses > 1:
                raise TypeError("Multiple XMLType classes %r matched for "
                                "for SAML AttributeValue type %r" % 
                                (XMLTypeClasses, input)) 
                   
            XMLTypeClass = XMLTypeClasses[0]            
        else:
            raise TypeError("Expecting %r class got %r" % (AttributeValue, 
                                                           type(input)))
        return XMLTypeClass
    

class IssuerElementTree(Issuer):
    """Represent a SAML Issuer element in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, issuer):
        """Create an XML representation of the input SAML issuer object
        
        @type issuer: saml.saml2.core.Issuer
        @param issuer: Assertion object
        @rtype: ElementTree.Element
        @return: ElementTree element containing the assertion
        """
        if not isinstance(issuer, Issuer):
            raise TypeError("Expecting %r class got %r" % (Issuer, 
                                                           type(issuer)))
            
        # Issuer format may be omitted from a response: saml-profiles-2.0-os,
        # Section 4.1.4.2
        attrib = {}
        if issuer.format is not None:
            attrib[cls.FORMAT_ATTRIB_NAME] = issuer.format
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        ElementTree._namespace_map[issuer.qname.namespaceURI
                                   ] = issuer.qname.prefix
                                   
        elem.text = issuer.value

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML Issuer instance
        
        @type elem: ElementTree.Element
        @param elem: ElementTree element containing the assertion
        @rtype: saml.saml2.core.Issuer
        @return: Assertion object"""
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError('No "%s" element found' %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        issuerFormat = elem.attrib.get(cls.FORMAT_ATTRIB_NAME)
        issuer = Issuer()
        
        # Issuer format may be omitted from a response: saml-profiles-2.0-os,
        # Section 4.1.4.2
        if issuerFormat is not None:
            issuer.format = issuerFormat
        
        if elem.text is None:
            raise XMLTypeParseError('No SAML issuer value set')
        
        issuer.value = elem.text.strip() 
        
        return issuer

        
class NameIdElementTree(NameID):
    """Represent a SAML Name Identifier in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, nameID):
        """Create an XML representation of the input SAML Name Identifier
        object
        @type nameID: saml.saml2.core.NameID
        @param nameID: SAML name ID
        @rtype: ElementTree.Element
        @return: Name ID as ElementTree XML element
        """
        
        if not isinstance(nameID, NameID):
            raise TypeError("Expecting %r class got %r" % (NameID, 
                                                           type(nameID)))
        attrib = {
            cls.FORMAT_ATTRIB_NAME: nameID.format
        }
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[nameID.qname.namespaceURI
                                   ] = nameID.qname.prefix
        
        elem.text = nameID.value

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML NameID object
        
        @type elem: ElementTree.Element
        @param elem: Name ID as ElementTree XML element
        @rtype: saml.saml2.core.NameID
        @return: SAML Name ID
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        format = elem.attrib.get(cls.FORMAT_ATTRIB_NAME)
        if format is None:
            raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                    'element' %
                                    (cls.FORMAT_ATTRIB_NAME,
                                     cls.DEFAULT_ELEMENT_LOCAL_NAME))
        nameID = NameID()
        nameID.format = format
        if elem.text is None:
            nameID.value = ''
        else:
            nameID.value = elem.text.strip() 
        
        return nameID


class SubjectElementTree(Subject):
    """Represent a SAML Subject in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, subject):
        """Create an XML representation of the input SAML subject object
        @type subject: saml.saml2.core.Subject
        @param subject: SAML subject
        @rtype: ElementTree.Element
        @return: subject as ElementTree XML element
        """
        if not isinstance(subject, Subject):
            raise TypeError("Expecting %r class got %r" % (Subject, 
                                                           type(subject)))
            
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))  
        elem = ElementTree.Element(tag)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix

            
        nameIdElem = NameIdElementTree.toXML(subject.nameID)
        elem.append(nameIdElem)
        
        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML Subject object
        
        @type elem: ElementTree.Element
        @param elem: subject as ElementTree XML element
        @rtype: saml.saml2.core.Subject
        @return: SAML subject
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        if len(elem) != 1:
            raise XMLTypeParseError("Expecting single Name ID child element "
                                      "for SAML Subject element")
            
        subject = Subject()
        subject.nameID = NameIdElementTree.fromXML(elem[0])
        
        return subject

        
class StatusCodeElementTree(StatusCode):
    """Represent a SAML Status Code in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, statusCode):
        """Create an XML representation of the input SAML Name Status Code
        
        @type statusCode: saml.saml2.core.StatusCode
        @param statusCode: SAML Status Code
        @rtype: ElementTree.Element
        @return: Status Code as ElementTree XML element
        """
        
        if not isinstance(statusCode, StatusCode):
            raise TypeError("Expecting %r class got %r" % (StatusCode, 
                                                           type(statusCode)))
            
        attrib = {
            cls.VALUE_ATTRIB_NAME: statusCode.value
        }
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[statusCode.qname.namespaceURI
                                   ] = statusCode.qname.prefix

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML StatusCode object
        
        @type elem: ElementTree.Element
        @param elem: Status Code as ElementTree XML element
        @rtype: saml.saml2.core.StatusCode
        @return: SAML Status Code
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError('No "%s" element found' %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        statusCode = StatusCode()
            
        value = elem.attrib.get(cls.VALUE_ATTRIB_NAME)
        if value is None:
            raise XMLTypeParseError('No "%s" attribute found in "%s" element' %
                                    (cls.VALUE_ATTRIB_NAME,
                                     cls.DEFAULT_ELEMENT_LOCAL_NAME))
        statusCode.value = value
        
        return statusCode

        
class StatusMessageElementTree(StatusMessage):
    """Represent a SAML Status Message in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, statusMessage):
        """Create an XML representation of the input SAML Name Status Message
        
        @type statusMessage: saml.saml2.core.StatusMessage
        @param statusMessage: SAML Status Message
        @rtype: ElementTree.Element
        @return: Status Code as ElementTree XML element
        """
        
        if not isinstance(statusMessage, StatusMessage):
            raise TypeError("Expecting %r class got %r" % (StatusMessage, 
                                                           type(statusMessage)))
            
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag)
        
        ElementTree._namespace_map[statusMessage.qname.namespaceURI
                                   ] = statusMessage.qname.prefix
        
        elem.text = statusMessage.value

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML StatusMessage object
        
        @type elem: ElementTree.Element
        @param elem: Status Code as ElementTree XML element
        @rtype: saml.saml2.core.StatusMessage
        @return: SAML Status Message
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError('No "%s" element found' %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        statusMessage = StatusMessage()
        if elem.text is not None:
            statusMessage.value = elem.text.strip() 
        
        return statusMessage


class StatusElementTree(Status):
    """Represent a SAML Status in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, status):
        """Create an XML representation of the input SAML subject object
        @type status: saml.saml2.core.Status
        @param status: SAML subject
        @rtype: ElementTree.Element
        @return: subject as ElementTree XML element
        """
        if not isinstance(status, Status):
            raise TypeError("Expecting %r class got %r" % (status, 
                                                           type(Status)))
            
        tag = str(QName.fromGeneric(Status.DEFAULT_ELEMENT_NAME))  
        elem = ElementTree.Element(tag)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix
        
        statusCodeElem = StatusCodeElementTree.toXML(status.statusCode)
        elem.append(statusCodeElem)
        
        # Status message is optional
        if status.statusMessage is not None and \
           status.statusMessage.value is not None:
            statusMessageElem = StatusMessageElementTree.toXML(
                                                        status.statusMessage)
            elem.append(statusMessageElem)
        
        if status.statusDetail is not None:
            raise NotImplementedError("StatusDetail XML serialisation is not "
                                      "implemented")
            
        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML Status object
        
        @type elem: ElementTree.Element
        @param elem: subject as ElementTree XML element
        @rtype: saml.saml2.core.Status
        @return: SAML subject
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != Status.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError('No "%s" element found' %
                                      Status.DEFAULT_ELEMENT_LOCAL_NAME)
            
        if len(elem) < 1:
            raise XMLTypeParseError("Expecting a StatusCode child element for "
                                    "SAML Status element")
            
        status = Status()
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName == StatusCode.DEFAULT_ELEMENT_LOCAL_NAME:
                status.statusCode = StatusCodeElementTree.fromXML(childElem)
                
            elif localName == StatusMessage.DEFAULT_ELEMENT_LOCAL_NAME:
                status.statusMessage = StatusMessageElementTree.fromXML(
                                                                childElem)
            elif localName == StatusDetail.DEFAULT_ELEMENT_LOCAL_NAME:
                raise NotImplementedError("XML parse of %s element is not "
                                    "implemented" %
                                    StatusDetail.DEFAULT_ELEMENT_LOCAL_NAME)
            else:
                raise XMLTypeParseError("Status child element name %r not "
                                        "recognised" % localName)
       
        return status
    
    
class AttributeQueryElementTree(AttributeQuery):
    """Represent a SAML Attribute Query in XML using ElementTree"""
        
    @classmethod
    def toXML(cls, attributeQuery, **attributeValueElementTreeFactoryKw):
        """Create an XML representation of the input SAML Attribute Query
        object

        @type attributeQuery: saml.saml2.core.AttributeQuery
        @param attributeQuery: SAML Attribute Query
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: ElementTree.Element
        @return: Attribute Query as ElementTree XML element
        """
        if not isinstance(attributeQuery, AttributeQuery):
            raise TypeError("Expecting %r class got %r" % (AttributeQuery, 
                                                        type(attributeQuery)))
            
        
        issueInstant = SAMLDateTime.toString(attributeQuery.issueInstant)
        attrib = {
            cls.ID_ATTRIB_NAME: attributeQuery.id,
            cls.ISSUE_INSTANT_ATTRIB_NAME: issueInstant,
            
            # Nb. Version is a SAMLVersion instance and requires explicit cast
            cls.VERSION_ATTRIB_NAME: str(attributeQuery.version)
        }
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix
        
        issuerElem = IssuerElementTree.toXML(attributeQuery.issuer)
        elem.append(issuerElem)

        subjectElem = SubjectElementTree.toXML(attributeQuery.subject)
        elem.append(subjectElem)

        for attribute in attributeQuery.attributes:
            # Factory enables support for multiple attribute types
            attributeElem = AttributeElementTree.toXML(attribute,
                                        **attributeValueElementTreeFactoryKw)
            elem.append(attributeElem)
        
        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML AttributeQuery object
        
        @type elem: ElementTree.Element
        @param elem: XML element containing the AttributeQuery
        @rtype: saml.saml2.core.AttributeQuery
        @return: AttributeQuery object
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        # Unpack attributes from top-level element
        attributeValues = []
        for attributeName in (cls.VERSION_ATTRIB_NAME,
                              cls.ISSUE_INSTANT_ATTRIB_NAME,
                              cls.ID_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                 'element' %
                                 (attributeName,
                                  cls.DEFAULT_ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
        
        attributeQuery = AttributeQuery()
        attributeQuery.version = SAMLVersion(attributeValues[0])
        if attributeQuery.version != SAMLVersion.VERSION_20:
            raise NotImplementedError("Parsing for %r is implemented for "
                                      "SAML version %s only; version %s is " 
                                      "not supported" % 
                                      (cls,
                                       SAMLVersion(SAMLVersion.VERSION_20),
                                       SAMLVersion(attributeQuery.version)))
            
        attributeQuery.issueInstant = SAMLDateTime.fromString(
                                                            attributeValues[1])
        attributeQuery.id = attributeValues[2]
        
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName == Issuer.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Issuer
                attributeQuery.issuer = IssuerElementTree.fromXML(childElem)
                
            elif localName == Subject.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Subject
                attributeQuery.subject = SubjectElementTree.fromXML(childElem)
            
            elif localName == Attribute.DEFAULT_ELEMENT_LOCAL_NAME:
                attribute = AttributeElementTree.fromXML(childElem)
                attributeQuery.attributes.append(attribute)
            else:
                raise XMLTypeParseError("Unrecognised AttributeQuery child "
                                          "element \"%s\"" % localName)
        
        return attributeQuery
        
    
class ResponseElementTree(Response):
    """Represent a SAML Response in XML using ElementTree"""
        
    @classmethod
    def toXML(cls, response, **attributeValueElementTreeFactoryKw):
        """Create an XML representation of the input SAML Response
        object

        @type response: saml.saml2.core.Response
        @param response: SAML Response
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        factory
        @rtype: ElementTree.Element
        @return: Response as ElementTree XML element
        """
        if not isinstance(response, Response):
            raise TypeError("Expecting %r class, got %r" % (Response, 
                                                            type(response)))
         
        if response.id is None:
            raise TypeError("SAML Response id is not set")
          
        if response.issueInstant is None:
            raise TypeError("SAML Response issueInstant is not set")
        
        # TODO: Does inResponseTo have to be set?  This implementation 
        # currently enforces this ...
        if response.inResponseTo is None:
            raise TypeError("SAML Response inResponseTo identifier is not set")
        
        issueInstant = SAMLDateTime.toString(response.issueInstant)
        attrib = {
            cls.ID_ATTRIB_NAME: response.id,
            cls.ISSUE_INSTANT_ATTRIB_NAME: issueInstant,
            cls.IN_RESPONSE_TO_ATTRIB_NAME: response.inResponseTo,
            
            # Nb. Version is a SAMLVersion instance and requires explicit cast
            cls.VERSION_ATTRIB_NAME: str(response.version)
        }
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))        
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix

        # Issuer may be omitted: saml-profiles-2.0-os Section 4.1.4.2
        if response.issuer is not None: 
            issuerElem = IssuerElementTree.toXML(response.issuer)
            elem.append(issuerElem)

        statusElem = StatusElementTree.toXML(response.status)       
        elem.append(statusElem)

        for assertion in response.assertions:
            # Factory enables support for multiple attribute types
            assertionElem = AssertionElementTree.toXML(assertion,
                                        **attributeValueElementTreeFactoryKw)
            elem.append(assertionElem)
        
        return elem

    @classmethod
    def fromXML(cls, elem, **attributeValueElementTreeFactoryKw):
        """Parse ElementTree element into a SAML Response object
        
        @type elem: ElementTree.Element
        @param elem: XML element containing the Response
        @type attributeValueElementTreeFactoryKw: dict
        @param attributeValueElementTreeFactoryKw: keywords for AttributeValue
        @rtype: saml.saml2.core.Response
        @return: Response object
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != Response.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                      Response.DEFAULT_ELEMENT_LOCAL_NAME)
        
        # Unpack attributes from top-level element
        attributeValues = []
        for attributeName in (Response.VERSION_ATTRIB_NAME,
                              Response.ISSUE_INSTANT_ATTRIB_NAME,
                              Response.ID_ATTRIB_NAME,
                              Response.IN_RESPONSE_TO_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                          'element' %
                                         (attributeName,
                                          Response.DEFAULT_ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
        
        response = Response()
        response.version = SAMLVersion(attributeValues[0])
        if response.version != SAMLVersion.VERSION_20:
            raise NotImplementedError("Parsing for %r is implemented for "
                                      "SAML version %s only; version %s is " 
                                      "not supported" % 
                                      (cls,
                                       SAMLVersion(SAMLVersion.VERSION_20),
                                       SAMLVersion(response.version)))
            
        response.issueInstant = SAMLDateTime.fromString(attributeValues[1])
        response.id = attributeValues[2]
        response.inResponseTo = attributeValues[3]
        
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName == Issuer.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Issuer
                response.issuer = IssuerElementTree.fromXML(childElem)
            
            elif localName == Status.DEFAULT_ELEMENT_LOCAL_NAME:
                # Get status of response
                response.status = StatusElementTree.fromXML(childElem)
                
            elif localName == Subject.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Subject
                response.subject = SubjectElementTree.fromXML(childElem)
            
            elif localName == Assertion.DEFAULT_ELEMENT_LOCAL_NAME:
                assertion = AssertionElementTree.fromXML(childElem,
                                        **attributeValueElementTreeFactoryKw)
                response.assertions.append(assertion)
            else:
                raise XMLTypeParseError('Unrecognised Response child '
                                          'element "%s"' % localName)
        
        return response


class ActionElementTree(Action):
    """Represent a SAML authorization action in XML using ElementTree"""
    
    @classmethod
    def toXML(cls, action):
        """Create an XML representation of the input SAML Name Identifier
        object
        @type action: saml.saml2.core.Action
        @param action: SAML subject
        @rtype: ElementTree.Element
        @return: Name ID as ElementTree XML element
        """
        
        if not isinstance(action, Action):
            raise TypeError("Expecting %r class got %r" % (Action, 
                                                           type(action)))
            
        if not action.namespace:
            raise AttributeError("No action namespace set")
        
        attrib = {
            cls.NAMESPACE_ATTRIB_NAME: action.namespace
        }
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[action.qname.namespaceURI
                                   ] = action.qname.prefix
        
        if not action.value:
            raise AttributeError("No action name set")
         
        elem.text = action.value

        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML Action object
        
        @type elem: ElementTree.Element
        @param elem: Name ID as ElementTree XML element
        @rtype: saml.saml2.core.Action
        @return: SAML Name ID
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
            
        action = Action()
        namespace = elem.attrib.get(cls.NAMESPACE_ATTRIB_NAME)
        if namespace is None:
            log.warning('No "%s" attribute found in "%s" element assuming '
                        '%r action namespace' %
                        (cls.NAMESPACE_ATTRIB_NAME,
                         cls.DEFAULT_ELEMENT_LOCAL_NAME,
                         action.namespace))
        else:
            action.namespace = namespace
            
        action.value = elem.text.strip() 
        
        return action
    
    
class AuthzDecisionQueryElementTree(AuthzDecisionQuery):
    """Represent a SAML Attribute Query in XML using ElementTree"""
        
    @classmethod
    def toXML(cls, authzDecisionQuery):
        """Create an XML representation of the input SAML Authorization
        Decision Query object

        @type authzDecisionQuery: saml.saml2.core.AuthzDecisionQuery
        @param authzDecisionQuery: SAML Authorization Decision Query
        @rtype: ElementTree.Element
        @return: Attribute Query as ElementTree XML element
        """
        if not isinstance(authzDecisionQuery, AuthzDecisionQuery):
            raise TypeError("Expecting %r class got %r" % (AuthzDecisionQuery, 
                                                    type(authzDecisionQuery)))
            
        if not authzDecisionQuery.resource:
            raise AttributeError("No resource has been set for the "
                                 "AuthzDecisionQuery")
            
        issueInstant = SAMLDateTime.toString(authzDecisionQuery.issueInstant)
        attrib = {
            cls.ID_ATTRIB_NAME: authzDecisionQuery.id,
            cls.ISSUE_INSTANT_ATTRIB_NAME: issueInstant,
            
            # Nb. Version is a SAMLVersion instance and requires explicit cast
            cls.VERSION_ATTRIB_NAME: str(authzDecisionQuery.version),
            
            cls.RESOURCE_ATTRIB_NAME: authzDecisionQuery.resource
        }
        
        tag = str(QName.fromGeneric(cls.DEFAULT_ELEMENT_NAME))
        elem = ElementTree.Element(tag, **attrib)
        
        ElementTree._namespace_map[cls.DEFAULT_ELEMENT_NAME.namespaceURI
                                   ] = cls.DEFAULT_ELEMENT_NAME.prefix
        
        issuerElem = IssuerElementTree.toXML(authzDecisionQuery.issuer)
        elem.append(issuerElem)

        subjectElem = SubjectElementTree.toXML(authzDecisionQuery.subject)
        elem.append(subjectElem)

        for action in authzDecisionQuery.actions:
            # Factory enables support for multiple attribute types
            actionElem = ActionElementTree.toXML(action)
            elem.append(actionElem)
        
        if (authzDecisionQuery.evidence and 
            len(authzDecisionQuery.evidence.evidence) > 0):
            raise NotImplementedError("Conversion of AuthzDecisionQuery "
                                      "Evidence type to ElementTree Element is "
                                      "not currently supported")
            
        return elem

    @classmethod
    def fromXML(cls, elem):
        """Parse ElementTree element into a SAML AuthzDecisionQuery object
        
        @type elem: ElementTree.Element
        @param elem: XML element containing the AuthzDecisionQuery
        @rtype: saml.saml2.core.AuthzDecisionQuery
        @return: AuthzDecisionQuery object
        """
        if not ElementTree.iselement(elem):
            raise TypeError("Expecting %r input type for parsing; got %r" %
                            (ElementTree.Element, elem))

        if QName.getLocalPart(elem.tag) != cls.DEFAULT_ELEMENT_LOCAL_NAME:
            raise XMLTypeParseError("No \"%s\" element found" %
                                    cls.DEFAULT_ELEMENT_LOCAL_NAME)
        
        # Unpack attributes from top-level element
        attributeValues = []
        for attributeName in (cls.VERSION_ATTRIB_NAME,
                              cls.ISSUE_INSTANT_ATTRIB_NAME,
                              cls.ID_ATTRIB_NAME,
                              cls.RESOURCE_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLTypeParseError('No "%s" attribute found in "%s" '
                                 'element' %
                                 (attributeName,
                                  cls.DEFAULT_ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
        
        authzDecisionQuery = AuthzDecisionQuery()
        authzDecisionQuery.version = SAMLVersion(attributeValues[0])
        if authzDecisionQuery.version != SAMLVersion.VERSION_20:
            raise NotImplementedError("Parsing for %r is implemented for "
                                      "SAML version %s only; version %s is " 
                                      "not supported" % 
                                      (cls,
                                       SAMLVersion(SAMLVersion.VERSION_20),
                                       SAMLVersion(authzDecisionQuery.version)))
            
        authzDecisionQuery.issueInstant = SAMLDateTime.fromString(
                                                            attributeValues[1])
        authzDecisionQuery.id = attributeValues[2]        
        authzDecisionQuery.resource = attributeValues[3]
        
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            if localName == Issuer.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Issuer
                authzDecisionQuery.issuer = IssuerElementTree.fromXML(childElem)
                
            elif localName == Subject.DEFAULT_ELEMENT_LOCAL_NAME:
                # Parse Subject
                authzDecisionQuery.subject = SubjectElementTree.fromXML(childElem)
            
            elif localName == Action.DEFAULT_ELEMENT_LOCAL_NAME:
                action = ActionElementTree.fromXML(childElem)
                authzDecisionQuery.actions.append(action)
            else:
                raise XMLTypeParseError("Unrecognised AuthzDecisionQuery child "
                                        "element \"%s\"" % localName)
        
        return authzDecisionQuery




"""NDG SOAP server package

NERC DataGrid Project
"""
__author__ = "Philip Kershaw"
__date__ = "26/01/2010"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: $'



"""SOAP Server helper module for unit test

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "27/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: soap_server.py 7130 2010-06-30 13:33:07Z pjkersha $"
from ndg.soap.test.test_soap import SOAPBindingMiddleware
from paste.httpserver import serve

if __name__ == "__main__":
    app = SOAPBindingMiddleware()
    serve(app, host='0.0.0.0', port=10080)




#!/usr/bin/env python
"""SOAP module unit test module

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "24/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: test_soap.py 7134 2010-06-30 13:49:40Z pjkersha $"
import logging
logging.basicConfig(level=logging.DEBUG)

import unittest
import socket
from cStringIO import StringIO
from os import path
import paste.fixture
from urllib2 import HTTPHandler, URLError

from ndg.soap.etree import SOAPEnvelope
from ndg.soap.client import UrlLib2SOAPClient, UrlLib2SOAPRequest
from ndg.soap.test import PasteDeployAppServer


class SOAPBindingMiddleware(object):
    """Simple WSGI interface for SOAP service"""
        
    def __call__(self, environ, start_response):
        requestFile = environ['wsgi.input']
        
        print("Server received request from client:\n\n%s" % 
              requestFile.read())
        
        soapResponse = SOAPEnvelope()
        soapResponse.create()
        
        response = soapResponse.serialize()
        start_response("200 OK",
                       [('Content-length', str(len(response))),
                        ('Content-type', 'text/xml')])
        return [response]
    
    
class SOAPTestCase(unittest.TestCase):
    SOAP_SERVICE_PORTNUM = 10080
    ENDPOINT = 'http://localhost:%d/soap' % SOAP_SERVICE_PORTNUM
    THIS_DIR = path.abspath(path.dirname(__file__))
    
    def __init__(self, *args, **kwargs):
        """Use paste.fixture to test client/server SOAP interface"""
        self.services = []
        self.disableServiceStartup = False

        wsgiApp = SOAPBindingMiddleware()
        self.app = paste.fixture.TestApp(wsgiApp)
         
        super(SOAPTestCase, self).__init__(*args, **kwargs)
        
    def test01Envelope(self):
        envelope = SOAPEnvelope()
        envelope.create()
        soap = envelope.serialize()
        
        self.assert_(len(soap) > 0)
        self.assert_("Envelope" in soap)
        self.assert_("Body" in soap)
        self.assert_("Header" in soap)
        
        print(envelope.prettyPrint())
        stream = StringIO()
        stream.write(soap)
        stream.seek(0)
        
        envelope2 = SOAPEnvelope()
        envelope2.parse(stream)
        soap2 = envelope2.serialize()
        self.assert_(soap2 == soap)

    def test02SendRequest(self):
        requestEnvelope = SOAPEnvelope()
        requestEnvelope.create()
        request = requestEnvelope.serialize()
        
        response = self.app.post('/my-soap-endpoint', 
                                 params=request, 
                                 status=200)
        print(response.headers)
        print(response.status)
        print(response.body)

    def test03Urllib2Client(self):
        
        # Paster based service is threaded from this call
        self.addService(app=SOAPBindingMiddleware(), 
                        port=SOAPTestCase.SOAP_SERVICE_PORTNUM)
        
        client = UrlLib2SOAPClient()
        
        # ElementTree based envelope class
        client.responseEnvelopeClass = SOAPEnvelope
        
        request = UrlLib2SOAPRequest()
        request.url = SOAPTestCase.ENDPOINT
        request.envelope = SOAPEnvelope()
        request.envelope.create()
        
        client.openerDirector.add_handler(HTTPHandler())
        try:
            response = client.send(request)
        except URLError, e:
            self.fail("soap_server.py must be running for this test")
        
        print("Response from server:\n\n%s" % response.envelope.serialize())
        
    def addService(self, *arg, **kw):
        """Utility for setting up threads to run Paste HTTP based services with
        unit tests
        
        @param arg: tuple contains ini file path setting for the service
        @type arg: tuple
        @param kw: keywords including "port" - port number to run the service 
        from
        @type kw: dict
        """
        if self.disableServiceStartup:
            return
        
        try:
            self.services.append(PasteDeployAppServer(*arg, **kw))
            self.services[-1].startThread()
            
        except socket.error:
            pass

    def __del__(self):
        """Stop any services started with the addService method and clean up
        the CA directory following the trust roots call
        """
        if hasattr(self, 'services'):
            for service in self.services:
                service.terminateThread()


if __name__ == "__main__":
    unittest.main()


"""SAML 2.0 Binding package

Implementation of SAML 2.0 for NDG Security

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $"



#!/usr/bin/env python
"""SAML 2.0 Package

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, 
Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "10/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: setup.py 7562 2010-10-01 12:52:44Z pjkersha $'

# Bootstrap setuptools if necessary.
from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup, find_packages
   
_longDescription = """\
SAML 2.0 implementation for use with the NERC DataGrid / Earth System Grid 
Project Attribute and Authorisation Query interfaces.  The implementation is 
based on the Java OpenSAML libraries.  An implementation is provided with  
ElementTree but it can easily be extended to use other Python XML parsers.

Version 0.5 adds WSGI middleware and clients for SAML SOAP binding and assertion
query/request profile.

It is not a complete implementation of SAML 2.0.  Only those components required
for the NERC DataGrid have been provided (Attribute and AuthZ Decision Query/
Response).  Where possible, stubs have been provided for other classes.
"""

setup(
    name =           		'ndg_saml',
    version =        		'0.5',
    description =    		('SAML 2.0 implementation for the NERC DataGrid '
                             'based on the Java OpenSAML library'),
    long_description =		_longDescription,
    author =         		'Philip Kershaw',
    author_email =   		'Philip.Kershaw@stfc.ac.uk',
    maintainer =         	'Philip Kershaw',
    maintainer_email =   	'Philip.Kershaw@stfc.ac.uk',
    url =            		'http://proj.badc.rl.ac.uk/ndg/wiki/Security/SAML2.0',
    license =               'http://www.apache.org/licenses/LICENSE-2.0',
    packages =			    find_packages(),
    namespace_packages =	['ndg'],
    extras_require = {
        'soap_binding':  ["M2Crypto", "PyOpenSSL", "Paste", "PasteDeploy", 
                          "PasteScript"],
        'zsi_soap_middleware': ['ZSI'],
    },
    include_package_data =  True,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Security',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe =              False
)



"""SOAP module unit tests

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "24/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $"
import logging
logging.basicConfig(level=logging.DEBUG)

import paste.httpserver
from threading import Thread
from paste.deploy import loadapp
from paste.script.util.logging_config import fileConfig


class PasteDeployAppServer(object):
    """Wrapper to paste.httpserver to enable background threading"""
    
    def __init__(self, app=None, cfgFilePath=None, port=7443, host='0.0.0.0',
                 ssl_context=None, withLoggingConfig=True):
        """Load an application configuration from cfgFilePath ini file and 
        instantiate Paste server object
        """       
        self.__thread = None
        
        if cfgFilePath:
            if withLoggingConfig:
                fileConfig(cfgFilePath)
            app = loadapp('config:%s' % cfgFilePath)
            
        elif app is None:
            raise KeyError('Either the "cfgFilePath" or "app" keyword must be '
                           'set')
                       
        self.__pasteServer = paste.httpserver.serve(app, host=host, port=port, 
                                                    start_loop=False, 
                                                    ssl_context=ssl_context)
    
    @property
    def pasteServer(self):
        return self.__pasteServer
    
    @property
    def thread(self):
        return self.__thread
    
    def start(self):
        """Start server"""
        self.pasteServer.serve_forever()
        
    def startThread(self):
        """Start server in a separate thread"""
        self.__thread = Thread(target=PasteDeployAppServer.start, args=(self,))
        self.thread.start()
        
    def terminateThread(self):
        self.pasteServer.server_close()



"""NDG SAML SOAP Binding unit test package

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7154 2010-07-01 15:59:59Z pjkersha $'



"""SAML 2.0 client bindings module implements SOAP binding for attribute query

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "02/09/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: attributequery.py 7322 2010-08-13 14:17:22Z pjkersha $'
import re
import logging
log = logging.getLogger(__name__)

from M2Crypto.m2urllib2 import HTTPSHandler

from ndg.saml.saml2.core import Attribute, AttributeQuery

from ndg.saml.utils import TypedList
from ndg.saml.saml2.binding.soap.client.subjectquery import (
                                                    SubjectQuerySOAPBinding,
                                                    SubjectQueryResponseError)

# Prevent whole module breaking if this is not available - it's only needed for
# AttributeQuerySslSOAPBinding
try:
    from ndg.saml.utils.m2crypto import SSLContextProxy
    _sslContextProxySupport = True
    
except ImportError:
    _sslContextProxySupport = False


class AttributeQueryResponseError(SubjectQueryResponseError):
    """SAML Response error from Attribute Query"""
    

class AttributeQuerySOAPBinding(SubjectQuerySOAPBinding): 
    """SAML Attribute Query SOAP Binding
    """
    QUERY_ATTRIBUTES_ATTRNAME = 'queryAttributes'
    LEN_QUERY_ATTRIBUTES_ATTRNAME = len(QUERY_ATTRIBUTES_ATTRNAME)
    QUERY_ATTRIBUTES_PAT = re.compile(',\s*')
    
    __PRIVATE_ATTR_PREFIX = "__"
    __slots__ = ()

    SERIALISE_KW = 'serialise'
    DESERIALISE_KW = 'deserialise'
    QUERY_TYPE = AttributeQuery
    
    def __init__(self, **kw):
        '''Create SOAP Client for SAML Attribute Query'''
        
        # Default to ElementTree based serialisation/deserialisation
        if AttributeQuerySOAPBinding.SERIALISE_KW not in kw:
            from ndg.saml.xml.etree import AttributeQueryElementTree
            kw[AttributeQuerySOAPBinding.SERIALISE_KW
               ] = AttributeQueryElementTree.toXML
               
        if AttributeQuerySOAPBinding.DESERIALISE_KW not in kw:
            from ndg.saml.xml.etree import ResponseElementTree
            kw[AttributeQuerySOAPBinding.DESERIALISE_KW
               ] = ResponseElementTree.fromXML
        
        super(AttributeQuerySOAPBinding, self).__init__(**kw)
            
    def __setattr__(self, name, value):
        """Enable setting of SAML query attribute objects via a comma separated
        string suitable for use reading from an ini file.  
        """
        try:
            super(AttributeQuerySOAPBinding, self).__setattr__(name, value)
            
        except AttributeError:
            if name.startswith(
                        AttributeQuerySOAPBinding.QUERY_ATTRIBUTES_ATTRNAME):
                # Special handler for parsing string format settings
                if not isinstance(value, basestring):
                    raise TypeError('Expecting string format for special '
                                    '%r attribute; got %r instead' %
                                    (name, type(value)))
                    
                pat = AttributeQuerySOAPBinding.QUERY_ATTRIBUTES_PAT
                attribute = Attribute()
                
                (attribute.name, 
                 attribute.friendlyName, 
                 attribute.nameFormat) = pat.split(value)
                 
                self.queryAttributes.append(attribute)
            else:
                raise
             
    def _getQueryAttributes(self):
        return self.query.attributes

    def _setQueryAttributes(self, value):
        if not isinstance(value, TypedList) and value.elementType != Attribute:
            raise TypeError('Expecting TypedList(Attribute) type for '
                            '"queryAttributes"; got %r instead' % type(value))
        
        # Remove all previously set items and add new ones 
        del self.query.attributes[:]
        for attribute in value:
            self.query.attributes.append(attribute)
  
    queryAttributes = property(_getQueryAttributes, 
                               _setQueryAttributes, 
                               doc="List of attributes to query from the "
                                   "Attribute Authority")

    
class AttributeQuerySslSOAPBinding(AttributeQuerySOAPBinding):
    """Specialisation of AttributeQuerySOAPbinding taking in the setting of
    SSL parameters for mutual authentication
    """
    SSL_CONTEXT_PROXY_SUPPORT = _sslContextProxySupport
    __slots__ = ('__sslCtxProxy',)
    
    def __init__(self, **kw):
        if not AttributeQuerySslSOAPBinding.SSL_CONTEXT_PROXY_SUPPORT:
            raise ImportError("ndg.saml.utils.m2crypto import "
                              "failed - missing M2Crypto package?")
        
        # Miss out default HTTPSHandler and set in send() instead
        if 'handlers' in kw:
            raise TypeError("__init__() got an unexpected keyword argument "
                            "'handlers'")
            
        super(AttributeQuerySslSOAPBinding, self).__init__(handlers=(), **kw)
        self.__sslCtxProxy = SSLContextProxy()

    def send(self, **kw):
        """Override base class implementation to pass explicit SSL Context
        """
        httpsHandler = HTTPSHandler(ssl_context=self.sslCtxProxy.createCtx())
        self.client.openerDirector.add_handler(httpsHandler)
        return super(AttributeQuerySslSOAPBinding, self).send(**kw)
            
    def _getSslCtxProxy(self):
        return self.__sslCtxProxy
    
    def _setSslCtxProxy(self, value):
        if not isinstance(value, SSLContextProxy):
            raise TypeError('Expecting %r type for "sslCtxProxy attribute; got '
                            '%r' % type(value))
            
        self.__sslCtxProxy = value
            
    sslCtxProxy = property(fget=_getSslCtxProxy, fset=_setSslCtxProxy,
                           doc="SSL Context Proxy object used for setting up "
                               "an SSL Context for queries")
    
    def __setattr__(self, name, value):
        """Enable setting of SSLContextProxy attributes as if they were 
        attributes of this class.  This is intended as a convenience for 
        making settings parameters read from a config file
        """
        try:
            super(AttributeQuerySslSOAPBinding, self).__setattr__(name, value)
            
        except AttributeError, e:
            # Coerce into setting SSL Context Proxy attributes
            try:
                setattr(self.sslCtxProxy, name, value)
            except:
                raise e



"""Implementation of SAML 2.0 for NDG Security - XML package

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "23/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $"
import logging
log = logging.getLogger(__name__)
   

class XMLConstants(object):
    '''XML related constants.
    
    @cvar XMLTOOLING_CONFIG_NS: Configuration namespace
    @type XMLTOOLING_CONFIG_NS: string
    @cvar XMLTOOLING_CONFIG_PREFIX: Configuration namespace prefix
    @type XMLTOOLING_CONFIG_PREFIX: string
    @cvar XMLTOOLING_DEFAULT_OBJECT_PROVIDER: object provider
    @type XMLTOOLING_DEFAULT_OBJECT_PROVIDER: string
    @cvar XML_NS: XML core namespace
    @type XML_NS: string
    @cvar XML_PREFIX: XML core prefix for xml attributes
    @type XML_PREFIX: string
    @cvar XMLNS_NS: XML namespace for xmlns attributes
    @type XMLNS_NS: string
    @cvar XMLNS_PREFIX: XML namespace prefix for xmlns attributes
    @type XMLNS_PREFIX: string
    @cvar XSD_NS: XML Schema namespace
    @type XSD_NS: string
    @cvar XSD_PREFIX: XML Schema QName prefix
    @type XSD_PREFIX: string
    @cvar XSI_NS: XML Schema Instance namespace
    @type XSI_NS: string
    @cvar XSI_PREFIX: XML Schema Instance QName prefix
    @type XSI_PREFIX: string
    @cvar XMLSIG_NS: XML XMLSecSignatureImpl namespace
    @type XMLSIG_NS: string
    @cvar XMLSIG_PREFIX: XML XMLSecSignatureImpl QName prefix
    @type XMLSIG_PREFIX: string
    @cvar XMLENC_NS: XML Encryption namespace
    @type XMLENC_NS: string
    @cvar XMLENC_PREFIX: XML Encryption QName prefix
    @type XMLENC_PREFIX: string
    @cvar XMLENC_ENCDATA_LOCAL_NAME: Local name of EncryptedData element
    @type XMLENC_ENCDATA_LOCAL_NAME: string
    @cvar XMLENC_ENCKEY_LOCAL_NAME: Local name of EncryptedKey element
    @type XMLENC_ENCKEY_LOCAL_NAME: string
    '''

    # XML Tooling

    # Configuration namespace
    XMLTOOLING_CONFIG_NS = "http://www.opensaml.org/xmltooling-config"

    # Configuration namespace prefix
    XMLTOOLING_CONFIG_PREFIX = "xt"
    
    # Name of the object provider used for objects that don't have a registered
    # object provider
    XMLTOOLING_DEFAULT_OBJECT_PROVIDER = "DEFAULT"

    # Core XML

    # XML core namespace
    XML_NS = "http://www.w3.org/XML/1998/namespace"
    
    # XML core prefix for xml attributes
    XML_PREFIX = "xml"

    # XML namespace for xmlns attributes
    XMLNS_NS = "http://www.w3.org/2000/xmlns/"

    # XML namespace prefix for xmlns attributes
    XMLNS_PREFIX = "xmlns"

    # XML Schema namespace
    XSD_NS = "http://www.w3.org/2001/XMLSchema"

    # XML Schema QName prefix
    XSD_PREFIX = "xs"

    # XML Schema Instance namespace
    XSI_NS = "http://www.w3.org/2001/XMLSchema-instance"

    # XML Schema Instance QName prefix
    XSI_PREFIX = "xsi"

    # XML XMLSecSignatureImpl namespace
    XMLSIG_NS = "http://www.w3.org/2000/09/xmldsig#"

    # XML XMLSecSignatureImpl QName prefix
    XMLSIG_PREFIX = "ds"

    # XML Encryption namespace
    XMLENC_NS = "http://www.w3.org/2001/04/xmlenc#"

    # XML Encryption QName prefix
    XMLENC_PREFIX = "xenc"
    
    # Local name of EncryptedData element
    XMLENC_ENCDATA_LOCAL_NAME = "EncryptedData"
    
    # Local name of EncryptedKey element
    XMLENC_ENCKEY_LOCAL_NAME = "EncryptedKey"


class XMLTypeError(Exception):
    """Generic XML type exception"""

class XMLTypeParseError(XMLTypeError):
    """Parse error for XML type"""

class UnknownAttrProfile(XMLTypeError):
    """Raise from Attribute Value factory if attribute type is not recognised
    """


"""WSGI SOAP Middleware utilising the ZSI Python SOAP Package

NERC DataGrid Project

"""
__author__ = "P J Kershaw"
__date__ = "19/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__revision__ = "$Id: $"
import logging
log = logging.getLogger(__name__)

import sys

from ZSI.parse import ParsedSoap
from ZSI.writer import SoapWriter
from ZSI import fault
from ZSI.ServiceContainer import ServiceSOAPBinding

from ndg.security.server.wsgi.soap import SOAPMiddleware, SOAPMiddlewareError,\
    SOAPMiddlewareConfigError, SOAPMiddlewareReadError
from ndg.security.common.wssecurity.utils import DomletteReader, \
    DomletteElementProxy
from ndg.security.common.utils.classfactory import instantiateClass, \
    importClass
      
class ZSIMiddlewareError(SOAPMiddlewareError):
    """Base class for ZSI Middleware type exceptions"""
    
class ZSIMiddlewareReadError(SOAPMiddlewareReadError):
    """ZSI Middleware read error"""

class ZSIMiddlewareConfigError(SOAPMiddlewareConfigError):
    """ZSI middleware configuration error"""
     
class ZSIMiddleware(SOAPMiddleware):
    '''Middleware configurable to a given ZSI SOAP binding

     @type SOAP_WRITER_KEYNAME: basestring
     @cvar SOAP_WRITER_KEYNAME: environ key for ZSI SoapWriter instance
     @type PARSED_SOAP_KEYNAME: basestring
     @cvar PARSED_SOAP_KEYNAME: environ key for ZSI ParsedSoap instance
     @type CHARSET_OPTNAME: basestring
     @cvar CHARSET_OPTNAME: option name to for character set for output
     @type DEFAULT_CHARSET: basestring
     @cvar DEFAULT_CHARSET: default character setting is utf-8
     @type PATH_OPTNAME: basestring
     @cvar PATH_OPTNAME: option to set path for this endpoint (not including 
     domain name)
     @type WRITE_RESPONSE_OPTNAME: basestring 
     @cvar WRITE_RESPONSE_OPTNAME: option name for flag to middleware to 
     serialise and output the SoapWriter instance
     @type REFERENCED_FILTERS_OPTNAME: basestring
     @cvar REFERENCED_FILTERS_OPTNAME: name for option to enable dereferencing 
     of other middleware via these environ keys
     @type FILTER_ID_OPTNAME: basestring
     @cvar FILTER_ID_OPTNAME: option name for environ key to enable other 
     middleware to reference this Filter
     @type PUBLISHED_URI_OPTNAME: basestring
     @cvar PUBLISHED_URI_OPTNAME: option name to define path for this endpoint
     including domain name
     @type READER_CLASS_OPTNAME: basestring
     @cvar READER_CLASS_OPTNAME: option name for SOAP reader class
     @type WRITERCLASS_OPTNAME: basestring
     @cvar WRITERCLASS_OPTNAME: option name for SOAP writer class
     '''  
    
    SOAP_WRITER_KEYNAME = 'ZSI.writer.SoapWriter'
    PARSED_SOAP_KEYNAME = 'ZSI.parse.ParsedSoap'
    
    CHARSET_OPTNAME = 'charset'
    DEFAULT_CHARSET = '; charset=utf-8'
    PATH_OPTNAME = 'path'
    WRITE_RESPONSE_OPTNAME = 'writeResponse'
    REFERENCED_FILTERS_OPTNAME = 'referencedFilters'
    FILTER_ID_OPTNAME = 'filterID'
    PUBLISHED_URI_OPTNAME = 'publishedURI'
    READER_CLASS_OPTNAME = 'readerclass'
    WRITERCLASS_OPTNAME = 'writerclass'
    
    def __init__(self, app):
        log.debug("ZSIMiddleware.__init__ ...")
        super(ZSIMiddleware, self).__init__()
        
        self._app = app
        self.__charset = ZSIMiddleware.DEFAULT_CHARSET
        self.__path = None
        self.__referencedFilterKeys = None
        self.__publishedURI = None
        self.__readerClass = None
        self.__writerClass = None
        self.__writeResponseSet = None
        self.__filterID = None

    def _getCharset(self):
        return self.__charset

    def _setCharset(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "charset" got %r' %
                            type(value))
        self.__charset = value

    def _getPath(self):
        return self.__path

    def _setPath(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "path" got %r' %
                            type(value))
        self.__path = value

    def _getPublishedURI(self):
        return self.__publishedURI

    def _setPublishedURI(self, value):
        if not isinstance(value, (basestring, type(None))):
            raise TypeError('Expecting string or None type for "publishedURI" '
                            'got %r' % type(value))
        self.__publishedURI = value

    def _getReaderClass(self):
        return self.__readerClass

    def _setReaderClass(self, value):
        self.__readerClass = value

    def _getWriterClass(self):
        return self.__writerClass

    def _setWriterClass(self, value):
        self.__writerClass = value

    def _getWriteResponseSet(self):
        return self.__writeResponseSet

    def _setWriteResponseSet(self, value):
        if not isinstance(value, bool):
            raise TypeError('Expecting %r for "writeResponseSet" type got %r' %
                            (bool, type(value)))
        
        self.__writeResponseSet = value

    def _getFilterID(self):
        return self.__filterID

    def _setFilterID(self, value):
        if not isinstance(value, (basestring, type(None))):
            raise TypeError('Expecting string or None type for "filterID" got '
                            '%r' % type(value))
        self.__filterID = value

    charset = property(_getCharset, _setCharset, 
                       doc="character set for response")

    path = property(_getPath, _setPath, doc="Path for endpoint")

    publishedURI = property(_getPublishedURI, _setPublishedURI, 
                            doc="fully qualified path for endpoint")

    readerClass = property(_getReaderClass, _setReaderClass, 
                           doc="ZSI Reader class")

    writeResponseSet = property(_getWriteResponseSet, _setWriteResponseSet, 
                                doc="boolean set to True to write out a "
                                    "response from this middleware")

    writerClass = property(_getWriterClass, _setWriterClass, 
                           doc="ZSI Writer Class")

    filterID = property(_getFilterID, _setFilterID, 
                        doc="enable the instance of this middleware to be "
                            "referenced in environ by this identifier")

    def initialise(self, global_conf, prefix='', **app_conf):
        """Set-up ZSI middleware interface attributes.  Overloaded base class 
        method to enable custom settings from app_conf
        
        @type global_conf: dict        
        @param global_conf: PasteDeploy global configuration dictionary
        @type prefix: basestring
        @param prefix: prefix for configuration items
        @type app_conf: dict        
        @param app_conf: PasteDeploy application specific configuration 
        dictionary
        """
        charsetOptName = prefix + ZSIMiddleware.CHARSET_OPTNAME
        if charsetOptName in app_conf:
            self.charset = '; charset=' + app_conf[charsetOptName]
        else:
            self.charset = '; charset=utf-8'
            
        pathOptName = prefix + ZSIMiddleware.PATH_OPTNAME
        if pathOptName in app_conf:
            if app_conf[pathOptName] != '/':
                self.path = app_conf[pathOptName].rstrip('/')
            else:
                self.path = app_conf[pathOptName]
        else:
            self.path = '/'

        # This flag if set to True causes this handler to call the 
        # start_response method and output the SOAP response
        writeResponseOptName = prefix + ZSIMiddleware.WRITE_RESPONSE_OPTNAME
        self.writeResponseSet = ZSIMiddleware.str2Bool(app_conf.get(
                                                    writeResponseOptName, ''))

        # Check for a list of other filters to be referenced by this one
        referencedFiltersOptName = prefix + \
                                    ZSIMiddleware.REFERENCED_FILTERS_OPTNAME
        if referencedFiltersOptName in app_conf:
            # __call__  may reference any filters in environ keyed by these
            # keywords
            self.referencedFilterKeys = app_conf.pop(
                                            referencedFiltersOptName).split()

        
        filterIdOptName = prefix + ZSIMiddleware.FILTER_ID_OPTNAME
        self.filterID = app_conf.pop(filterIdOptName, None)
        
        # The endpoint that this services will be referenced from externally.
        # e.g. the Session Manager client running locally can check the
        # input URI and compare with this value to see if the request is 
        # actually to the local Session Manager instance
        publishedUriOptName = prefix + ZSIMiddleware.PUBLISHED_URI_OPTNAME
        self.publishedURI = app_conf.pop(publishedUriOptName, None)
        
        readerClassOptName = prefix + ZSIMiddleware.READER_CLASS_OPTNAME
        if readerClassOptName in app_conf:
            readerClassName = app_conf.pop(readerClassOptName)
            self.readerClass = importClass(readerClassName)
        else:
            self.readerClass = DomletteReader
            
        writerClassOptName = prefix + ZSIMiddleware.WRITERCLASS_OPTNAME
        if writerClassOptName in app_conf:
            writerClassName = app_conf.pop(writerClassOptName)
            self.writerClass = importClass(writerClassName)
        else:
            self.writerClass = DomletteElementProxy

    def __call__(self, environ, start_response):
        log.debug("ZSIMiddleware.__call__")
                        
        # Derived class must implement SOAP Response via overloaded version of
        # this method.  ParsedSoap object is available as a key in environ via
        # the parseRequest method
        
        return self.writeResponse(environ, start_response)

    
    def _initCall(self, environ, start_response):
        '''Sub-divided out from __call__ to enable derived classes to easily
        include this functionality:
         - Set a reference to this WSGI filter in environ if filterID was 
        set in the config and 
         - check the request to see if this filter should handle it
        '''
        
        # Add any filter references for this WSGI component regardless of the
        # current request ID.  This ensures that other WSGI components called
        # may reference it if they need to.
        self.addFilter2Environ(environ)
        
        # Apply filter for calls
        if not self.__class__.isSOAPMessage(environ):
            log.debug("ZSIMiddleware.__call__: skipping non-SOAP call")
            return self._app(environ, start_response)
        
        elif not self.pathMatch(environ):
            log.debug("ZSIMiddleware.__call__: path doesn't match SOAP "
                      "service endpoint")
            return self._app(environ, start_response)
        
        elif self.__class__.isSOAPFaultSet(environ):
            # This MUST be checked in a overloaded version especially in 
            # consideration of security: e.g. an upstream signature 
            # verification may have found an error in a signature
            log.debug("ZSIMiddleware.__call__: SOAP fault set by previous "
                      "SOAP middleware filter")
            return self._app(environ, start_response)

        # Parse input into a ZSI ParsedSoap object set as a key in environ
        try:
            self.parseRequest(environ)
        except Exception, e:
            sw = self.exception2SOAPFault(environ, e)
            self.setSOAPWriter(environ, sw)
            return self.writeResponse(environ, start_response)
        
        # Return None to __call__ to indicate that it can proceed with 
        # processing the input
        return None

    def exception2SOAPFault(self, environ, exception):
        '''Convert an exception into a SOAP fault message'''
        soapFault = fault.FaultFromException(exception, 
                                             None,
                                             tb=sys.exc_info()[2])
        sw = SoapWriter(outputclass=self.writerClass)
        soapFault.serialize(sw)
        environ[ZSIMiddleware.SOAP_FAULT_SET_KEYNAME] = True
        return sw
    
    pathMatch = lambda self, environ: environ['PATH_INFO'] == self.path
    
    def parseRequest(self, environ):
        '''Parse SOAP message from environ['wsgi.input']
        
        Reading from environ['wsgi.input'] may be a destructive process so the
        content is saved in a ZSI.parse.ParsedSoap object for use by SOAP
        handlers which follow in the chain
        
        environ['ZSI.parse.ParsedSoap'] may be set to a ParsedSoap object
        parsed by a SOAP handler ahead of the current one in the chain.  In
        this case, don't re-parse.  If NOT parsed, parse and set
        'ZSI.parse.ParsedSoap' environ key'''
        
        # Check for ParsedSoap object set in environment, if not present,
        # make one
        ps = environ.get(ZSIMiddleware.PARSED_SOAP_KEYNAME)
        if ps is None:
            # TODO: allow for chunked data
            contentLength = int(environ['CONTENT_LENGTH'])
            soapIn = environ['wsgi.input'].read(contentLength)
            if len(soapIn) < contentLength:
                raise ZSIMiddlewareReadError("Expecting %d content length; "
                                             "received %d instead." % 
                                             (contentLength, len(soapIn)))
            
            log.debug("SOAP Request for handler %r" % ZSIMiddleware)
            log.debug("_" * 80)
            log.debug(soapIn)
            log.debug("_" * 80)
            
            ps = ParsedSoap(soapIn, readerclass=self.readerClass)
            environ[ZSIMiddleware.PARSED_SOAP_KEYNAME] = ps
            
        return environ[ZSIMiddleware.PARSED_SOAP_KEYNAME]


    def writeResponse(self, environ, start_response, errorCode=None):
        '''This method serializes the SOAP output and sets the response header.
        It's the final step and should be called in the last SOAP handler in 
        a chain of handlers or else specify it in the ini file as the last
        SOAP handler'''
        
        # This flag must be set to True to write out the final response from
        # this handler
        if self.writeResponseSet == False:
            return self._app(environ, start_response)
        
        sw = self.getSOAPWriter(environ)
        soapOut = str(sw)
        
        if errorCode is None:
            if self.__class__.isSOAPFaultSet(environ):
                errorCode = "500 Internal Server Error"
            else:
                errorCode = "200 OK"
                
        log.debug("SOAP Response for handler %r" % self.__class__)
        log.debug("_" * 80)
        log.debug(soapOut)
        log.debug("_" * 80)
        start_response(errorCode,
                       [('content-type', 'text/xml' + self.charset),
                        ('content-length', str(len(soapOut)))])
        return soapOut

    @classmethod
    def getSOAPWriter(cls, environ):
        '''Access SoapWriter object set in environment by this classes' call
        method'''
        
        sw = environ.get(ZSIMiddleware.SOAP_WRITER_KEYNAME)
        if sw is None:
            raise KeyError("Expecting '%s' key in environ: missing call to "
                           "ZSIMiddleware?" % ZSIMiddleware.SOAP_WRITER_KEYNAME)
        return sw
    
    @classmethod
    def setSOAPWriter(cls, environ, sw):
        '''Set SoapWriter object in environment'''   
        environ[ZSIMiddleware.SOAP_WRITER_KEYNAME] = sw

    def addFilter2Environ(self, environ):
        '''Add a key to the current application in the environment so that
        other middleware can reference it.  This is dependent on filterID set
        in app_conf'''
        if self.filterID is not None:           
            if self.filterID in environ:
                raise ZSIMiddlewareConfigError("An filterID key '%s' is "
                                                "already set in environ" % 
                                                self.filterID)
            environ[self.filterID] = self


class SOAPBindingMiddleware(ZSIMiddleware):  
    '''Interface to apply a ZSI ServiceSOAPBinding type SOAP service'''
    
    SERVICE_SOAP_BINDING_CLASSNAME_OPTNAME = 'serviceSOAPBindingClass'
    SERVICE_SOAP_BINDING_PROPPREFIX_OPTNAME = 'serviceSOAPBindingPropPrefix'
    DEFAULT_SERVICE_SOAP_BINDING_PROPPREFIX_OPTNAME = \
                            'ndg.security.server.wsgi.zsi.serviceSOAPBinding.'
                            
    SERVICE_SOAP_BINDING_ENVIRON_KEYNAME_OPTNAME = \
                            'serviceSOAPBindingEnvironKeyName' 
    DEFAULT_SERVICE_SOAP_BINDING_ENVIRON_KEYNAME = \
                            'ndg.security.servier.wsgi.zsi.serviceSOAPBinding'
    ENABLE_WSDL_QUERY_OPTNAME = 'enableWSDLQuery' 
    DEFAULT_ENABLE_WSDL_QUERY_VALUE = False
    SOAP_METHOD_STRING = 'soap_%s'
    
    def __init__(self, app):
        super(SOAPBindingMiddleware, self).__init__(app)
        self.__serviceSOAPBindingKeyName = None
        self.__serviceSOAPBinding = None
        self.__enableWSDLQuery = False
        
    def _getServiceSOAPBinding(self):
        return self.__serviceSOAPBinding

    def _setServiceSOAPBinding(self, value):
        """Instance must be ZSI ServiceSOAPBinding derived type"""
        if not isinstance(value, ServiceSOAPBinding):
            raise TypeError('Expecting %r type for "serviceSOAPBinding"; got '
                            '%r' % (ServiceSOAPBinding, type(value)))
        self.__serviceSOAPBinding = value

    serviceSOAPBinding = property(fget=_getServiceSOAPBinding, 
                                  fset=_setServiceSOAPBinding, 
                                  doc="Instance of ZSI ServiceSOAPBinding "
                                      "derived type determines the behaviour "
                                      "of the SOAP callbacks")

    def _getServiceSOAPBindingKeyName(self):
        return self.__serviceSOAPBindingKeyName

    def _setServiceSOAPBindingKeyName(self, value):
        """Instance must be ZSI ServiceSOAPBindingKeyName derived type"""
        if not isinstance(value, basestring):
            raise TypeError('Expecting bool type for "enableWSDLQuery"; got '
                            '%r' % type(value))
        self.__serviceSOAPBindingKeyName = value

    serviceSOAPBindingKeyName = property(fget=_getServiceSOAPBindingKeyName, 
                                         fset=_setServiceSOAPBindingKeyName, 
                                         doc="Keyword to key WSGI environ for "
                                             "SOAP Binding middleware instance")

    def _getEnableWSDLQuery(self):
        return self.__enableWSDLQuery

    def _setEnableWSDLQuery(self, value):
        if not isinstance(value, bool):
            raise TypeError('Expecting bool type for "enableWSDLQuery"; got '
                            '%r' % type(value))
        self.__enableWSDLQuery = value
        
    enableWSDLQuery = property(fget=_getEnableWSDLQuery, 
                               fset=_setEnableWSDLQuery, 
                               doc="Enable service to publish the WSDL via "
                                   "the ?wsdl query argument appended to the "
                                   "endpoint")

    def initialise(self, global_conf, prefix='', **app_conf):
        """Set-up ZSI SOAP Binding middleware interface attributes.
         Overloaded base class method to enable custom settings from app_conf
        
        @type global_conf: dict        
        @param global_conf: PasteDeploy global configuration dictionary
        @type prefix: basestring
        @param prefix: prefix for configuration items
        @type app_conf: dict        
        @param app_conf: PasteDeploy application specific configuration 
        dictionary
        """
        super(SOAPBindingMiddleware, self).initialise(global_conf, 
                                                      prefix=prefix,
                                                      **app_conf)
        
        serviceSOAPBindingEnvironKeyNameOptName = prefix + \
            SOAPBindingMiddleware.SERVICE_SOAP_BINDING_ENVIRON_KEYNAME_OPTNAME
        serviceSOAPBindingClassNameOptName = prefix + \
            SOAPBindingMiddleware.SERVICE_SOAP_BINDING_CLASSNAME_OPTNAME
                            
        if serviceSOAPBindingEnvironKeyNameOptName in app_conf and \
           serviceSOAPBindingClassNameOptName in app_conf:
            raise ZSIMiddlewareConfigError('Only "%s" or "%s" may be set; not '
                                           ' both' % 
            (SOAPBindingMiddleware.SERVICE_SOAP_BINDING_ENVIRON_KEYNAME_OPTNAME,
             SOAPBindingMiddleware.SERVICE_SOAP_BINDING_CLASSNAME_OPTNAME))
            
        if serviceSOAPBindingClassNameOptName in app_conf:
            # Instantiate the binding from the config settings
            modName, className = app_conf[
                            serviceSOAPBindingClassNameOptName].rsplit('.', 1)
                
            # Filter
            prefixOptName = prefix + \
                SOAPBindingMiddleware.SERVICE_SOAP_BINDING_PROPPREFIX_OPTNAME
            prefix = app_conf.get(prefixOptName,
                            SOAPBindingMiddleware.
                            DEFAULT_SERVICE_SOAP_BINDING_PROPPREFIX_OPTNAME)
            
            serviceSOAPBindingKw = dict([(k.replace(prefix, ''), v)
                                         for k, v in app_conf.items()
                                         if k.startswith(prefix)])
    
            self.serviceSOAPBinding = instantiateClass(modName,
                                       className,
                                       objectType=ServiceSOAPBinding,
                                       classProperties=serviceSOAPBindingKw)           
        else: 
            # Alternatively, reference another binding instance made available
            # by upstream middleware via environ
            self.serviceSOAPBindingKeyName = app_conf.get(
                            serviceSOAPBindingEnvironKeyNameOptName,
                            SOAPBindingMiddleware.
                            DEFAULT_SERVICE_SOAP_BINDING_ENVIRON_KEYNAME)
        
        
        # Flag to enable display of WSDL via wsdl query arg in a GET request
        enableWSDLQueryOptName = prefix + \
                                SOAPBindingMiddleware.ENABLE_WSDL_QUERY_OPTNAME
        self.enableWSDLQuery = SOAPBindingMiddleware.str2Bool(
                                    app_conf.get(enableWSDLQueryOptName, ''))   
    
    def __call__(self, environ, start_response):
        log.debug("SOAPBindingMiddleware.__call__ ...")
             
        # Get a reference to the service SOAP binding from environ or if not,
        # from the binding instantiated at initialisation
        serviceSOAPBinding = environ.get(self.serviceSOAPBindingKeyName,
                                         self.serviceSOAPBinding)
        if serviceSOAPBinding is None:
            raise ZSIMiddlewareConfigError('No SOAP service binding set in '
                                           'environ or configured from start'
                                           '-up')
             
        if self.pathMatch(environ) and self.enableWSDLQuery and \
           environ.get('REQUEST_METHOD', '') == 'GET' and \
           environ.get('QUERY_STRING', '') == 'wsdl':
            wsdl = serviceSOAPBinding._wsdl
            start_response("200 OK", [('Content-type', 'text/xml'),
                                      ('Content-length', str(len(wsdl)))])
            return wsdl 
                
        # Apply filter for calls
        response = self._initCall(environ, start_response)
        if response is not None:
            return response
          
        try:
            # Other filters in the middleware chain may be passed by setting
            # a reference to them in the config.  This is useful if the SOAP
            # binding code needs to access results from upstream middleware 
            # e.g. check output from signature verification filter
            if hasattr(self, 'referencedFilterKeys'):
                try:
                    serviceSOAPBinding.referencedWSGIFilters = \
                                    dict([(i, environ[i]) 
                                          for i in self.referencedFilterKeys])
                except KeyError:
                    raise ZSIMiddlewareConfigError('No filter ID "%s" found '
                                                   'in environ' % i)    
            ps = self.parseRequest(environ)
               
            # Map SOAP Action to method in binding class
            soapActionName = environ[
                            SOAPBindingMiddleware.SOAP_ACTION_ENVIRON_KEYNAME
                            ].strip('"')
            soapMethodName = SOAPBindingMiddleware.SOAP_METHOD_STRING % \
                            soapActionName
                    
            
            method = getattr(serviceSOAPBinding, soapMethodName)            
            resp = method(ps)
        except Exception, e:
            sw = self.exception2SOAPFault(environ, e)
        else: 
            # Serialize output using SOAP writer class
            sw = SoapWriter(outputclass=self.writerClass)
            sw.serialize(resp)
        
        # Make SoapWriter object available to any SOAP filters that follow
        self.setSOAPWriter(environ, sw)
        soapOut = str(sw)

        return self.writeResponse(environ, start_response)



"""SAML Generic SOAP Binding Query/Response Interface unit test module

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "21/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: test_queryresponseinterface.py 7526 2010-09-27 12:08:12Z pjkersha $'
import logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)
import unittest

from datetime import datetime, timedelta
import os
from uuid import uuid4
import paste.fixture
from cStringIO import StringIO
from xml.etree import ElementTree

from ndg.saml.utils import SAMLDateTime
from ndg.saml.saml2.core import (Response, Assertion, Attribute, 
                             AttributeStatement, SAMLVersion, Subject, NameID,
                             Issuer, AttributeQuery, XSStringAttributeValue, 
                             Conditions, Status, StatusCode)
from ndg.saml.xml import XMLConstants
from ndg.saml.xml.etree import AttributeQueryElementTree, ResponseElementTree
from ndg.saml.saml2.binding.soap.client.subjectquery import (
    SubjectQuerySOAPBinding, ResponseIssueInstantInvalid, 
    AssertionIssueInstantInvalid, AssertionConditionNotBeforeInvalid, 
    AssertionConditionNotOnOrAfterInvalid)

from ndg.soap.client import (UrlLib2SOAPClient, UrlLib2SOAPRequest)
from ndg.soap.etree import SOAPEnvelope
from ndg.soap.utils.etree import QName, prettyPrint


class SamlSoapBindingApp(object):
    """Simple WSGI application to handle SAML Attribute Query/Response
    """
    FIRSTNAME_ATTRNAME = "urn:ndg:saml:firstname"
    LASTNAME_ATTRNAME = "urn:ndg:saml:lastname"
    EMAILADDRESS_ATTRNAME = "urn:ndg:saml:emailaddress"
    NAMEID_FORMAT = "urn:ndg:saml:openid"
    
    def __init__(self):
        self.firstName = "Philip"
        self.lastName = "Kershaw"
        self.emailAddress = "pkershaw@somewhere.ac.uk"
                  
    def __call__(self, environ, start_response):
        soapRequestStream = environ['wsgi.input']
        soapRequest = SOAPEnvelope()
        soapRequest.parse(soapRequestStream)
        attributeQueryElem = soapRequest.body.elem[0]
        attributeQuery = AttributeQueryElementTree.fromXML(attributeQueryElem)
        
        print("Received request from client:\n")
        print soapRequest.prettyPrint()
        
        samlResponse = Response()
        
        samlResponse.issueInstant = datetime.utcnow()
        samlResponse.id = str(uuid4())
        samlResponse.issuer = Issuer()
        
        # SAML 2.0 spec says format must be omitted
        #samlResponse.issuer.format = Issuer.X509_SUBJECT
        samlResponse.issuer.value = \
                        "/O=NDG/OU=BADC/CN=attributeauthority.badc.rl.ac.uk"
        
        samlResponse.inResponseTo = attributeQuery.id
        
        assertion = Assertion()
        
        assertion.version = SAMLVersion(SAMLVersion.VERSION_20)
        assertion.id = str(uuid4())
        assertion.issueInstant = samlResponse.issueInstant
        
        assertion.conditions = Conditions()
        assertion.conditions.notBefore = assertion.issueInstant
        assertion.conditions.notOnOrAfter = assertion.conditions.notBefore + \
            timedelta(seconds=60*60*8)
        
        assertion.subject = Subject()  
        assertion.subject.nameID = NameID()
        assertion.subject.nameID.format = attributeQuery.subject.nameID.format
        assertion.subject.nameID.value = attributeQuery.subject.nameID.value

        assertion.attributeStatements.append(AttributeStatement())
        
        for attribute in attributeQuery.attributes:
            if attribute.name == SamlSoapBindingApp.FIRSTNAME_ATTRNAME:
                # special case handling for 'FirstName' attribute
                fnAttribute = Attribute()
                fnAttribute.name = attribute.name
                fnAttribute.nameFormat = attribute.nameFormat
                fnAttribute.friendlyName = attribute.friendlyName
    
                firstName = XSStringAttributeValue()
                firstName.value = self.firstName
                fnAttribute.attributeValues.append(firstName)
    
                assertion.attributeStatements[0].attributes.append(fnAttribute)
            
            elif attribute.name == SamlSoapBindingApp.LASTNAME_ATTRNAME:
                lnAttribute = Attribute()
                lnAttribute.name = attribute.name
                lnAttribute.nameFormat = attribute.nameFormat
                lnAttribute.friendlyName = attribute.friendlyName
    
                lastName = XSStringAttributeValue()
                lastName.value = self.lastName
                lnAttribute.attributeValues.append(lastName)
    
                assertion.attributeStatements[0].attributes.append(lnAttribute)
               
            elif attribute.name == SamlSoapBindingApp.EMAILADDRESS_ATTRNAME:
                emailAddressAttribute = Attribute()
                emailAddressAttribute.name = attribute.name
                emailAddressAttribute.nameFormat = attribute.nameFormat
                emailAddressAttribute.friendlyName = attribute.friendlyName
    
                emailAddress = XSStringAttributeValue()
                emailAddress.value = self.emailAddress
                emailAddressAttribute.attributeValues.append(emailAddress)
    
                assertion.attributeStatements[0].attributes.append(
                                                        emailAddressAttribute)
        
        samlResponse.assertions.append(assertion)
        
        samlResponse.status = Status()
        samlResponse.status.statusCode = StatusCode()
        samlResponse.status.statusCode.value = StatusCode.SUCCESS_URI        

        
        # Convert to ElementTree representation to enable attachment to SOAP
        # response body
        samlResponseElem = ResponseElementTree.toXML(samlResponse)
        xml = ElementTree.tostring(samlResponseElem)
        log.debug('Sending response to query:\n%s', xml)
        
        # Create SOAP response and attach the SAML Response payload
        soapResponse = SOAPEnvelope()
        soapResponse.create()
        soapResponse.body.elem.append(samlResponseElem)
        
        response = soapResponse.serialize()
        
        start_response("200 OK",
                       [('Content-length', str(len(response))),
                        ('Content-type', 'text/xml')])
        return [response]

        
class SamlAttributeQueryTestCase(unittest.TestCase):
    """Test the SAML SOAP binding using an Attribute Query as an example"""
    thisDir = os.path.dirname(os.path.abspath(__file__))
    RESPONSE = '''\
<soap11:Envelope xmlns:soap11="http://schemas.xmlsoap.org/soap/envelope/">
   <soap11:Body>
      <samlp:Response ID="05680cb2-4973-443d-9d31-7bc99bea87c1" InResponseTo="e3183380-ae82-4285-8827-8c40613842de" IssueInstant="%(issueInstant)s" Version="2.0" xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol">
         <saml:Issuer Format="urn:esg:issuer" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">Somewhere</saml:Issuer>
         <samlp:Status>
            <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success" />
         </samlp:Status>
         <saml:Assertion ID="192c67d9-f9cd-457a-9242-999e7b943166" IssueInstant="%(assertionIssueInstant)s" Version="2.0" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
            <saml:Issuer Format="urn:esg:issuer">Somewhere</saml:Issuer>
            <saml:Subject>
               <saml:NameID Format="urn:esg:openid">https://somewhere.edu/myopenid/testUser</saml:NameID>
            </saml:Subject>
            <saml:Conditions NotBefore="%(notBefore)s" NotOnOrAfter="%(notOnOrAfter)s" />
            <saml:AttributeStatement>
               <saml:Attribute FriendlyName="FirstName" Name="urn:esg:first:name" NameFormat="http://www.w3.org/2001/XMLSchema#string">
                  <saml:AttributeValue xsi:type="xs:string" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">Test</saml:AttributeValue>
               </saml:Attribute>
               <saml:Attribute FriendlyName="LastName" Name="urn:esg:last:name" NameFormat="http://www.w3.org/2001/XMLSchema#string">
                  <saml:AttributeValue xsi:type="xs:string" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">User</saml:AttributeValue>
               </saml:Attribute>
               <saml:Attribute FriendlyName="EmailAddress" Name="urn:esg:first:email:address" NameFormat="http://www.w3.org/2001/XMLSchema#string">
                  <saml:AttributeValue xsi:type="xs:string" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">someone@somewhere.edu</saml:AttributeValue>
               </saml:Attribute>
            </saml:AttributeStatement>
         </saml:Assertion>
      </samlp:Response>
   </soap11:Body>
</soap11:Envelope>
'''

    def __init__(self, *args, **kwargs):
        wsgiApp = SamlSoapBindingApp()
        self.app = paste.fixture.TestApp(wsgiApp)
         
        unittest.TestCase.__init__(self, *args, **kwargs)
        
    def test01AttributeQuery(self):
        attributeQuery = AttributeQuery()
        attributeQuery.version = SAMLVersion(SAMLVersion.VERSION_20)
        attributeQuery.id = str(uuid4())
        attributeQuery.issueInstant = datetime.utcnow()
        
        attributeQuery.issuer = Issuer()
        attributeQuery.issuer.format = Issuer.X509_SUBJECT
        attributeQuery.issuer.value = \
                        "/O=NDG/OU=BADC/CN=attributeauthority.badc.rl.ac.uk"
                        
                        
        attributeQuery.subject = Subject()  
        attributeQuery.subject.nameID = NameID()
        attributeQuery.subject.nameID.format = SamlSoapBindingApp.NAMEID_FORMAT
        attributeQuery.subject.nameID.value = \
                                    "https://openid.localhost/philip.kershaw"
        
        # special case handling for 'FirstName' attribute
        fnAttribute = Attribute()
        fnAttribute.name = SamlSoapBindingApp.FIRSTNAME_ATTRNAME
        fnAttribute.nameFormat = "http://www.w3.org/2001/XMLSchema#string"
        fnAttribute.friendlyName = "FirstName"

        attributeQuery.attributes.append(fnAttribute)
    
        # special case handling for 'LastName' attribute
        lnAttribute = Attribute()
        lnAttribute.name = SamlSoapBindingApp.LASTNAME_ATTRNAME
        lnAttribute.nameFormat = "http://www.w3.org/2001/XMLSchema#string"
        lnAttribute.friendlyName = "LastName"

        attributeQuery.attributes.append(lnAttribute)
    
        # special case handling for 'LastName' attribute
        emailAddressAttribute = Attribute()
        emailAddressAttribute.name = SamlSoapBindingApp.EMAILADDRESS_ATTRNAME
        emailAddressAttribute.nameFormat = XMLConstants.XSD_NS+"#"+\
                                    XSStringAttributeValue.TYPE_LOCAL_NAME
        emailAddressAttribute.friendlyName = "emailAddress"

        attributeQuery.attributes.append(emailAddressAttribute)                                   
        
        elem = AttributeQueryElementTree.toXML(attributeQuery)
        soapRequest = SOAPEnvelope()
        soapRequest.create()
        soapRequest.body.elem.append(elem)
        
        request = soapRequest.serialize()
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        response = self.app.post('/attributeauthority', 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        print("Response status=%d" % response.status)

        soapResponse = SOAPEnvelope()
        
        responseStream = StringIO()
        responseStream.write(response.body)
        responseStream.seek(0)
        
        soapResponse.parse(responseStream)
        
        print("Parsed response ...")
        print(soapResponse.serialize())
#        print(prettyPrint(soapResponse.elem))
        
        response = ResponseElementTree.fromXML(soapResponse.body.elem[0])
        self.assert_(response.status.statusCode.value==StatusCode.SUCCESS_URI)
        self.assert_(response.inResponseTo == attributeQuery.id)
        self.assert_(response.assertions[0].subject.nameID.value == \
                     attributeQuery.subject.nameID.value)

    def _parseResponse(self, responseStr):
        """Helper to parse a response from a string"""
        soapResponse = SOAPEnvelope()
        
        responseStream = StringIO()
        responseStream.write(responseStr)
        responseStream.seek(0)
        
        soapResponse.parse(responseStream)
        
        print("Parsed response ...")
        print(soapResponse.serialize())
        
        response = ResponseElementTree.fromXML(soapResponse.body.elem[0])
        return response
        
    def test03ParseResponse(self):
        utcNow = datetime.utcnow()
        respDict = {
            'issueInstant': SAMLDateTime.toString(utcNow),
            'assertionIssueInstant': SAMLDateTime.toString(utcNow),
            'notBefore': SAMLDateTime.toString(utcNow),
            'notOnOrAfter': SAMLDateTime.toString(utcNow + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                        respDict
        response = self._parseResponse(responseStr)
        self.assert_(response)

    def test04AssertionConditionExpired(self):
        # issued 9 hours ago
        issueInstant = datetime.utcnow() - timedelta(seconds=60*60*9)
        respDict = {
            'issueInstant': SAMLDateTime.toString(issueInstant),
            'assertionIssueInstant': SAMLDateTime.toString(issueInstant),
            'notBefore': SAMLDateTime.toString(issueInstant),
            # It lasts for 8 hours so it's expired by one hour
            'notOnOrAfter': SAMLDateTime.toString(issueInstant + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                    respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        try:
            binding._verifyTimeConditions(response)
            self.fail("Expecting not on or after timestamp error")
        except AssertionConditionNotOnOrAfterInvalid, e:
            print("PASSED: %s" % e)

    def test05ResponseIssueInstantInvalid(self):
        utcNow = datetime.utcnow()
        respDict = {
            'issueInstant': SAMLDateTime.toString(utcNow + timedelta(
                                                                    seconds=1)),
            'assertionIssueInstant': SAMLDateTime.toString(utcNow),
            'notBefore': SAMLDateTime.toString(utcNow),
            'notOnOrAfter': SAMLDateTime.toString(utcNow + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                    respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        try:
            binding._verifyTimeConditions(response)
            self.fail("Expecting issue instant timestamp error")
        except ResponseIssueInstantInvalid, e:
            print("PASSED: %s" % e)

    def test06NotBeforeConditionInvalid(self):
        utcNow = datetime.utcnow()
        respDict = {
            'issueInstant': SAMLDateTime.toString(utcNow),
            'assertionIssueInstant': SAMLDateTime.toString(utcNow),
            'notBefore': SAMLDateTime.toString(utcNow + timedelta(seconds=1)),
            'notOnOrAfter': SAMLDateTime.toString(utcNow + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                    respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        try:
            binding._verifyTimeConditions(response)
            self.fail("Expecting issue instant timestamp error")
        except AssertionConditionNotBeforeInvalid, e:
            print("PASSED: %s" % e)

    def test07AssertionIssueInstantInvalid(self):
        utcNow = datetime.utcnow()
        respDict = {
            'issueInstant': SAMLDateTime.toString(utcNow),
            'assertionIssueInstant': SAMLDateTime.toString(utcNow + timedelta(
                                                                    seconds=1)),
            'notBefore': SAMLDateTime.toString(utcNow),
            'notOnOrAfter': SAMLDateTime.toString(utcNow + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                    respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        try:
            binding._verifyTimeConditions(response)
            self.fail("Expecting issue instant timestamp error")
        except AssertionIssueInstantInvalid, e:
            print("PASSED: %s" % e)

    def test07ClockSkewCorrectedAssertionIssueInstantInvalid(self):
        utcNow = datetime.utcnow()
        respDict = {
            'issueInstant': SAMLDateTime.toString(utcNow),
            'assertionIssueInstant': SAMLDateTime.toString(utcNow + timedelta(
                                                                    seconds=1)),
            'notBefore': SAMLDateTime.toString(utcNow),
            'notOnOrAfter': SAMLDateTime.toString(utcNow + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % \
                                                                    respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        
        # Set a skew to correct the error
        binding.clockSkewTolerance = 1
        
        try:
            binding._verifyTimeConditions(response)
        except AssertionIssueInstantInvalid, e:
            self.fail("issue instant timestamp error should be corrected for")

    def test08ClockSkewCorrectedAssertionConditionExpired(self):
        # Issued 9 hours ago
        issueInstant = datetime.utcnow() - timedelta(seconds=60*60*9)
        respDict = {
            'issueInstant': SAMLDateTime.toString(issueInstant),
            'assertionIssueInstant': SAMLDateTime.toString(issueInstant),
            'notBefore': SAMLDateTime.toString(issueInstant),
            # Assertion lasts 8 hours so it has expired by one hour
            'notOnOrAfter': SAMLDateTime.toString(issueInstant + timedelta(
                                                            seconds=60*60*8))
        }
        responseStr = self.__class__.RESPONSE % respDict
        response = self._parseResponse(responseStr)
        binding = SubjectQuerySOAPBinding()
        
        # Set a skew of over one hour to correct for the assertion expiry
        binding.clockSkewTolerance = 60*60 + 3
        
        try:
            binding._verifyTimeConditions(response)
            
        except AssertionConditionNotOnOrAfterInvalid, e:
            self.fail("Not on or after timestamp error should be corrected for")
            
            
if __name__ == "__main__":
    unittest.main()        

